#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Tests for elyra-pipeline application"""
import json
import os

from click.testing import CliRunner

import elyra.cli.pipeline_app as pipeline_app
from elyra.cli.pipeline_app import pipeline

SUB_COMMANDS = ['run', 'submit', 'describe', 'validate']

PIPELINE_SOURCE_WITH_ZERO_LENGTH_PIPELINES_FIELD = \
    '{"doc_type":"pipeline","version":"3.0","id":"0","primary_pipeline":"1","pipelines":[],"schemas":[]}'

PIPELINE_SOURCE_WITHOUT_PIPELINES_FIELD = \
    '{"doc_type":"pipeline","version":"3.0","id":"0","primary_pipeline":"1","schemas":[]}'

PIPELINE_SOURCE_WITH_ZERO_NODES = \
    '{"doc_type":"pipeline","version":"3.0","id":"0","primary_pipeline":"1","pipelines":[{"id":"1","nodes":[],"app_data":{"runtime":"","version": 5,"properties": {"name": "generic","runtime": "Generic"}}, "schemas":[]}]}'  # noqa


def mock_get_runtime_type(runtime_config: str) -> str:
    return "kfp"


def test_no_opts():
    runner = CliRunner()
    result = runner.invoke(pipeline)
    assert 'run       Run a pipeline in your local environment' in result.output
    assert 'submit    Submit a pipeline to be executed on the server' in result.output
    assert 'describe  Display pipeline summary' in result.output
    assert result.exit_code == 0


def test_bad_subcommand():
    runner = CliRunner()
    result = runner.invoke(pipeline, ['invalid_command'])
    assert "Error: No such command 'invalid_command'" in result.output
    assert result.exit_code != 0


def test_subcommand_no_opts():
    runner = CliRunner()
    for command in SUB_COMMANDS:
        result = runner.invoke(pipeline, [command])
        assert "Error: Missing argument 'PIPELINE_PATH'" in result.output
        assert result.exit_code != 0


def test_run_with_invalid_pipeline():
    runner = CliRunner()

    result = runner.invoke(pipeline, ['run', 'foo.pipeline'])
    assert "Pipeline file not found:" in result.output
    assert "foo.pipeline" in result.output
    assert result.exit_code != 0


def test_submit_with_invalid_pipeline(monkeypatch):
    runner = CliRunner()

    monkeypatch.setattr(pipeline_app, "_get_runtime_type", mock_get_runtime_type)

    result = runner.invoke(pipeline, ['submit', 'foo.pipeline',
                                      '--runtime-config', 'foo'])
    assert "Pipeline file not found:" in result.output
    assert "foo.pipeline" in result.output
    assert result.exit_code != 0


def test_describe_with_invalid_pipeline():
    runner = CliRunner()

    result = runner.invoke(pipeline, ['describe', 'foo.pipeline'])
    assert "Pipeline file not found:" in result.output
    assert "foo.pipeline" in result.output
    assert result.exit_code != 0


def test_validate_with_invalid_pipeline():
    runner = CliRunner()

    result = runner.invoke(pipeline, ['validate', 'foo.pipeline'])
    assert "Pipeline file not found:" in result.output
    assert "foo.pipeline" in result.output
    assert result.exit_code != 0


def test_run_with_unsupported_file_type():
    runner = CliRunner()
    with runner.isolated_filesystem():
        with open('foo.ipynb', 'w') as f:
            f.write('{ "nbformat": 4, "cells": [] }')

        result = runner.invoke(pipeline, ['run', 'foo.ipynb'])
        assert "Pipeline file should be a [.pipeline] file" in result.output
        assert result.exit_code != 0


def test_submit_with_unsupported_file_type(monkeypatch):
    runner = CliRunner()
    with runner.isolated_filesystem():
        with open('foo.ipynb', 'w') as f:
            f.write('{ "nbformat": 4, "cells": [] }')

        monkeypatch.setattr(pipeline_app, "_get_runtime_type", mock_get_runtime_type)

        result = runner.invoke(pipeline, ['submit', 'foo.ipynb',
                                          '--runtime-config', 'foo'])
        assert "Pipeline file should be a [.pipeline] file" in result.output
        assert result.exit_code != 0


def test_describe_with_unsupported_file_type():
    runner = CliRunner()
    with runner.isolated_filesystem():
        with open('foo.ipynb', 'w') as f:
            f.write('{ "nbformat": 4, "cells": [] }')

        result = runner.invoke(pipeline, ['describe', 'foo.ipynb'])
        assert "Pipeline file should be a [.pipeline] file" in result.output
        assert result.exit_code != 0


def test_validate_with_unsupported_file_type():
    runner = CliRunner()
    with runner.isolated_filesystem():
        with open('foo.ipynb', 'w') as f:
            f.write('{ "nbformat": 4, "cells": [] }')

        result = runner.invoke(pipeline, ['validate', 'foo.ipynb'])
        assert "Pipeline file should be a [.pipeline] file" in result.output
        assert result.exit_code != 0


def test_run_with_no_pipelines_field():
    runner = CliRunner()
    with runner.isolated_filesystem():
        with open('foo.pipeline', 'w') as pipeline_file:
            pipeline_file.write(PIPELINE_SOURCE_WITHOUT_PIPELINES_FIELD)
            pipeline_file_path = os.path.join(os.getcwd(), pipeline_file.name)

        result = runner.invoke(pipeline, ['run', pipeline_file_path])
        assert "Pipeline is missing 'pipelines' field." in result.output
        assert result.exit_code != 0


def test_submit_with_no_pipelines_field(monkeypatch):
    runner = CliRunner()
    with runner.isolated_filesystem():
        with open('foo.pipeline', 'w') as pipeline_file:
            pipeline_file.write(PIPELINE_SOURCE_WITHOUT_PIPELINES_FIELD)
            pipeline_file_path = os.path.join(os.getcwd(), pipeline_file.name)

        monkeypatch.setattr(pipeline_app, "_get_runtime_type", mock_get_runtime_type)

        result = runner.invoke(pipeline, ['submit', pipeline_file_path,
                                          '--runtime-config', 'foo'])
        assert "Pipeline is missing 'pipelines' field." in result.output
        assert result.exit_code != 0


def test_describe_with_no_pipelines_field():
    runner = CliRunner()
    with runner.isolated_filesystem():
        with open('foo.pipeline', 'w') as pipeline_file:
            pipeline_file.write(PIPELINE_SOURCE_WITHOUT_PIPELINES_FIELD)
            pipeline_file_path = os.path.join(os.getcwd(), pipeline_file.name)

        result = runner.invoke(pipeline, ['describe', pipeline_file_path])
        assert "Pipeline is missing 'pipelines' field." in result.output
        assert result.exit_code != 0


def test_validate_with_no_pipelines_field():
    runner = CliRunner()
    with runner.isolated_filesystem():
        with open('foo.pipeline', 'w') as pipeline_file:
            pipeline_file.write(PIPELINE_SOURCE_WITHOUT_PIPELINES_FIELD)
            pipeline_file_path = os.path.join(os.getcwd(), pipeline_file.name)

        result = runner.invoke(pipeline, ['validate', pipeline_file_path])
        assert "Pipeline is missing 'pipelines' field." in result.output
        assert result.exit_code != 0


def test_run_with_zero_length_pipelines_field():
    runner = CliRunner()
    with runner.isolated_filesystem():
        with open('foo.pipeline', 'w') as pipeline_file:
            pipeline_file.write(PIPELINE_SOURCE_WITH_ZERO_LENGTH_PIPELINES_FIELD)
            pipeline_file_path = os.path.join(os.getcwd(), pipeline_file.name)

        result = runner.invoke(pipeline, ['run', pipeline_file_path])
        assert "Pipeline has zero length 'pipelines' field." in result.output
        assert result.exit_code != 0


def test_submit_with_zero_length_pipelines_field(monkeypatch):
    runner = CliRunner()
    with runner.isolated_filesystem():
        with open('foo.pipeline', 'w') as pipeline_file:
            pipeline_file.write(PIPELINE_SOURCE_WITH_ZERO_LENGTH_PIPELINES_FIELD)
            pipeline_file_path = os.path.join(os.getcwd(), pipeline_file.name)

        monkeypatch.setattr(pipeline_app, "_get_runtime_type", mock_get_runtime_type)

        result = runner.invoke(pipeline, ['submit', pipeline_file_path,
                                          '--runtime-config', 'foo'])
        assert "Pipeline has zero length 'pipelines' field." in result.output
        assert result.exit_code != 0


def test_describe_with_zero_length_pipelines_field():
    runner = CliRunner()
    with runner.isolated_filesystem():
        with open('foo.pipeline', 'w') as pipeline_file:
            pipeline_file.write(PIPELINE_SOURCE_WITH_ZERO_LENGTH_PIPELINES_FIELD)
            pipeline_file_path = os.path.join(os.getcwd(), pipeline_file.name)

        result = runner.invoke(pipeline, ['describe', pipeline_file_path])
        assert "Pipeline has zero length 'pipelines' field." in result.output
        assert result.exit_code != 0


def test_run_pipeline_with_no_nodes():
    runner = CliRunner()
    with runner.isolated_filesystem():
        with open('foo.pipeline', 'w') as pipeline_file:
            pipeline_file.write(PIPELINE_SOURCE_WITH_ZERO_NODES)
            pipeline_file_path = os.path.join(os.getcwd(), pipeline_file.name)

        result = runner.invoke(pipeline, ['run', pipeline_file_path])
        assert "At least one node must exist in the primary pipeline." in result.output
        assert result.exit_code != 0


def test_submit_pipeline_with_no_nodes(monkeypatch):
    runner = CliRunner()
    with runner.isolated_filesystem():
        with open('foo.pipeline', 'w') as pipeline_file:
            pipeline_file.write(PIPELINE_SOURCE_WITH_ZERO_NODES)
            pipeline_file_path = os.path.join(os.getcwd(), pipeline_file.name)

        monkeypatch.setattr(pipeline_app, "_get_runtime_type", mock_get_runtime_type)

        result = runner.invoke(pipeline, ['submit', pipeline_file_path, '--runtime-config', 'foo'])
        assert "At least one node must exist in the primary pipeline." in result.output
        assert result.exit_code != 0


def test_describe_with_empty_pipeline():
    runner = CliRunner()
    with runner.isolated_filesystem():
        with open('foo.pipeline', 'w') as pipeline_file:
            pipeline_file.write(PIPELINE_SOURCE_WITH_ZERO_NODES)
            pipeline_file_path = os.path.join(os.getcwd(), pipeline_file.name)

        result = runner.invoke(pipeline, ['describe', pipeline_file_path])
        assert "Description: None" in result.output
        assert "Type: generic" in result.output
        assert "Nodes: 0" in result.output


def test_describe_with_kfp_components():
    runner = CliRunner()
    pipeline_file_path = os.path.join(os.path.dirname(__file__), 'resources', 'kfp_3_node_custom.pipeline')

    result = runner.invoke(pipeline, ['describe', pipeline_file_path])
    assert "Description: 3-node custom component pipeline" in result.output
    assert "Type: kfp" in result.output
    assert "Nodes: 3" in result.output
    assert result.exit_code == 0


def test_validate_with_kfp_components():
    runner = CliRunner()
    pipeline_file_path = os.path.join(os.path.dirname(__file__), 'resources', 'kfp_3_node_custom.pipeline')

    result = runner.invoke(pipeline, ['validate', pipeline_file_path])
    assert "Validating pipeline..." in result.output
    assert result.exit_code == 0


def test_describe_with_missing_kfp_component():
    runner = CliRunner()
    with runner.isolated_filesystem():
        valid_file_path = os.path.join(os.path.dirname(__file__), 'resources', 'kfp_3_node_custom.pipeline')
        pipeline_file_path = os.path.join(os.getcwd(), 'foo.pipeline')
        with open(pipeline_file_path, 'w') as pipeline_file:
            with open(valid_file_path) as valid_file:
                valid_data = json.load(valid_file)
                # Update known component name to trigger a missing component
                valid_data['pipelines'][0]['nodes'][0]['op'] = valid_data['pipelines'][0]['nodes'][0]['op'] + 'Missing'
                pipeline_file.write(json.dumps(valid_data))

        result = runner.invoke(pipeline, ['describe', pipeline_file_path])
        assert "Description: 3-node custom component pipeline" in result.output
        assert "Type: kfp" in result.output
        assert "Nodes: 3" in result.output
        assert result.exit_code == 0


def test_validate_with_missing_kfp_component():
    runner = CliRunner()
    with runner.isolated_filesystem():
        valid_file_path = os.path.join(os.path.dirname(__file__), 'resources', 'kfp_3_node_custom.pipeline')
        pipeline_file_path = os.path.join(os.getcwd(), 'foo.pipeline')
        with open(pipeline_file_path, 'w') as pipeline_file:
            with open(valid_file_path) as valid_file:
                valid_data = json.load(valid_file)
                # Update known component name to trigger a missing component
                valid_data['pipelines'][0]['nodes'][0]['op'] = valid_data['pipelines'][0]['nodes'][0]['op'] + 'Missing'
                pipeline_file.write(json.dumps(valid_data))

        result = runner.invoke(pipeline, ['validate', pipeline_file_path])
        assert "Validating pipeline..." in result.output
        assert "[Error][Calculate data hash] - This component was not found in the registry." in result.output
        assert result.exit_code != 0



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from elyra.pipeline.pipeline_definition import PipelineDefinition
from elyra.tests.pipeline.util import _read_pipeline_resource


def test_valid_pipeline():
    pipeline_json = _read_pipeline_resource('resources/sample_pipelines/pipeline_valid.json')

    pipeline_definition = PipelineDefinition(pipeline_definition=pipeline_json)

    assert pipeline_definition.is_valid()


def test_validation_flags_missing_schema_version():
    _check_missing_pipeline_field('version', 'Pipeline schema version field is missing.')


def test_validation_flags_schema_version_has_wrong_type():
    _check_pipeline_field_type('version', 3.0, "Pipeline schema version field should be a string.")


def test_validation_flags_missing_pipelines_field():
    _check_missing_pipeline_field('pipelines', "Pipeline is missing 'pipelines' field.")


def test_validation_flags_pipelines_has_wrong_type():
    _check_pipeline_field_type('pipelines', '', "Field 'pipelines' should be a list.")


def test_validation_flags_pipelines_is_empty():
    _check_pipeline_field_type('pipelines', list(), "Pipeline has zero length 'pipelines' field.")


def test_validation_flags_missing_primary_pipeline_field():
    _check_missing_pipeline_field('primary_pipeline', "Could not determine the primary pipeline.")


def test_validation_flags_missing_primary_pipeline_nodes_field():
    _check_missing_primary_pipeline_field('nodes', "At least one node must exist in the primary pipeline.")


def test_validation_flags_missing_app_data_field():
    _check_missing_primary_pipeline_field('app_data', "Primary pipeline is missing the 'app_data' field.")


def test_validation_flags_missing_version_field():
    pipeline_json = _read_pipeline_resource('resources/sample_pipelines/pipeline_valid.json')
    pipeline_json['pipelines'][0]['app_data'].pop('version')

    pipeline_definition = PipelineDefinition(pipeline_definition=pipeline_json)

    assert pipeline_definition.is_valid() is False
    assert "Primary pipeline is missing the 'version' field." in pipeline_definition.validate()


def test_updates_to_primary_pipeline_updates_pipeline_definition():
    pipeline_json = _read_pipeline_resource('resources/sample_pipelines/pipeline_valid.json')

    pipeline_definition = PipelineDefinition(pipeline_definition=pipeline_json)
    pipeline_definition.primary_pipeline.set('version', 3)

    assert pipeline_definition.primary_pipeline.version == 3
    assert pipeline_definition.to_dict()['pipelines'][0]['app_data']['version'] == 3


def test_updates_to_nodes_updates_pipeline_definition():
    pipeline_json = _read_pipeline_resource('resources/sample_pipelines/pipeline_valid.json')

    pipeline_definition = PipelineDefinition(pipeline_definition=pipeline_json)
    for node in pipeline_definition.primary_pipeline.nodes:
        node.set_component_parameter('filename', 'foo')

    for node in pipeline_definition.to_dict()['pipelines'][0]['nodes']:
        assert node['app_data']['component_parameters']['filename'] == 'foo'


def _check_pipeline_correct_pipeline_name():
    pipeline_json = _read_pipeline_resource('resources/sample_pipelines/pipeline_valid.json')
    pipeline_definition = PipelineDefinition(pipeline_definition=pipeline_json)

    primary_pipeline = pipeline_definition.primary_pipeline

    assert primary_pipeline.name == "{{name}}"


def _check_pipeline_correct_pipeline_alternative_name():
    pipeline_json = _read_pipeline_resource('resources/sample_pipelines/pipeline_valid_alternative_name.json')
    pipeline_definition = PipelineDefinition(pipeline_definition=pipeline_json)

    primary_pipeline = pipeline_definition.primary_pipeline

    assert primary_pipeline.name == "{{alternative_name}}"

# Utility questions


def _check_missing_pipeline_field(field: str, error_msg: str):
    pipeline_json = _read_pipeline_resource('resources/sample_pipelines/pipeline_valid.json')
    pipeline_json.pop(field)

    pipeline_definition = PipelineDefinition(pipeline_definition=pipeline_json)

    assert pipeline_definition.is_valid() is False
    assert error_msg in pipeline_definition.validate()


def _check_pipeline_field_type(field: str, wrong_type_value: any, error_msg: str):
    pipeline_json = _read_pipeline_resource('resources/sample_pipelines/pipeline_valid.json')
    pipeline_json.pop(field)
    pipeline_json[field] = wrong_type_value

    pipeline_definition = PipelineDefinition(pipeline_definition=pipeline_json)

    assert pipeline_definition.is_valid() is False
    assert error_msg in pipeline_definition.validate()


def _check_missing_primary_pipeline_field(field: str, error_msg: str):
    pipeline_json = _read_pipeline_resource('resources/sample_pipelines/pipeline_valid.json')
    pipeline_json['pipelines'][0].pop(field)

    pipeline_definition = PipelineDefinition(pipeline_definition=pipeline_json)

    assert pipeline_definition.is_valid() is False
    assert error_msg in pipeline_definition.validate()



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
from types import SimpleNamespace

import jupyter_core.paths

from elyra.metadata.manager import MetadataManager
from elyra.metadata.metadata import Metadata
from elyra.metadata.schemaspaces import ComponentRegistries
from elyra.pipeline.component import FilesystemComponentReader
from elyra.pipeline.component import UrlComponentReader
from elyra.pipeline.component_registry import ComponentRegistry
from elyra.pipeline.kfp.component_parser_kfp import KfpComponentParser

COMPONENT_CATALOG_DIRECTORY = os.path.join(jupyter_core.paths.ENV_JUPYTER_PATH[0], 'components')


def _get_resource_path(filename):
    root = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))
    resource_path = os.path.join(root, '..', '..', '..', 'tests/pipeline', 'resources', 'components', filename)
    resource_path = os.path.normpath(resource_path)

    return resource_path


def test_component_registry_can_load_components_from_registries():
    component_parser = KfpComponentParser()
    component_registry = ComponentRegistry(component_parser)

    components = component_registry.get_all_components()
    assert len(components) > 0


def test_modify_component_registries():
    # Get initial set of components from the current active registries
    parser = KfpComponentParser()
    component_registry = ComponentRegistry(parser, caching_enabled=False)
    initial_components = component_registry.get_all_components()

    # Components must be sorted by id for the equality comparison with later component lists
    initial_components = sorted(initial_components, key=lambda component: component.id)

    metadata_manager = MetadataManager(schemaspace=ComponentRegistries.COMPONENT_REGISTRIES_SCHEMASPACE_ID)

    # Create new registry instance with a single URL-based component
    paths = [_get_resource_path('kfp_test_operator.yaml')]

    instance_metadata = {
        "description": "A test registry",
        "runtime": "kfp",
        "categories": ["New Components"],
        "location_type": "Filename",
        "paths": paths
    }
    registry_instance = Metadata(schema_name="component-registry",
                                 name="new_registry",
                                 display_name="New Registry",
                                 metadata=instance_metadata)

    try:
        if metadata_manager.get("new_registry"):
            metadata_manager.remove("new_registry")
    except Exception:
        pass

    metadata_manager.create("new_registry", registry_instance)

    # Get new set of components from all active registries, including added test registry
    added_components = component_registry.get_all_components()
    assert len(added_components) > len(initial_components)

    added_component_names = [component.name for component in added_components]
    assert 'Test Operator' in added_component_names
    assert 'Test Operator No Inputs' not in added_component_names

    # Modify the test registry to add an additional path to
    paths.append(_get_resource_path('kfp_test_operator_no_inputs.yaml'))
    metadata_manager.update("new_registry", registry_instance)

    # Get set of components from all active registries, including modified test registry
    modified_components = component_registry.get_all_components()
    assert len(modified_components) > len(added_components)

    modified_component_names = [component.name for component in modified_components]
    assert 'Test Operator' in modified_component_names
    assert 'Test Operator No Inputs' in modified_component_names

    # Delete the test registry
    metadata_manager.remove("new_registry")
    post_delete_components = component_registry.get_all_components()
    post_delete_components = sorted(post_delete_components, key=lambda component: component.id)
    assert len(post_delete_components) == len(initial_components)

    # Check that the list of component ids is the same as before addition of the test registry
    initial_component_ids = [component.id for component in initial_components]
    post_delete_component_ids = [component.id for component in post_delete_components]
    assert post_delete_component_ids == initial_component_ids

    # Check that component palette is the same as before addition of the test registry
    initial_palette = ComponentRegistry.to_canvas_palette(post_delete_components)
    post_delete_palette = ComponentRegistry.to_canvas_palette(initial_components)
    assert initial_palette == post_delete_palette


def test_directory_based_component_registry():
    # Get initial set of components from the current active registries
    parser = KfpComponentParser()
    component_registry = ComponentRegistry(parser, caching_enabled=False)
    initial_components = component_registry.get_all_components()

    metadata_manager = MetadataManager(schemaspace=ComponentRegistries.COMPONENT_REGISTRIES_SCHEMASPACE_ID)

    # Create new directory-based registry instance with components in ../../test/resources/components
    registry_path = _get_resource_path('')
    instance_metadata = {
        "description": "A test registry",
        "runtime": "kfp",
        "categories": ["New Components"],
        "location_type": "Directory",
        "paths": [registry_path]
    }
    registry_instance = Metadata(schema_name="component-registry",
                                 name="new_registry",
                                 display_name="New Registry",
                                 metadata=instance_metadata)

    try:
        if metadata_manager.get("new_registry"):
            metadata_manager.remove("new_registry")
    except Exception:
        pass

    metadata_manager.create("new_registry", registry_instance)

    # Get new set of components from all active registries, including added test registry
    added_components = component_registry.get_all_components()
    assert len(added_components) > len(initial_components)

    # Check that all relevant components from the new registry have been added
    added_component_names = [component.name for component in added_components]
    assert 'Filter text' in added_component_names
    assert 'Test Operator' in added_component_names
    assert 'Test Operator No Inputs' in added_component_names

    # Remove the test instance
    metadata_manager.remove("new_registry")


def test_parse_kfp_component_file():
    # Define the appropriate reader for a filesystem-type component definition
    kfp_supported_file_types = [".yaml"]
    reader = FilesystemComponentReader(kfp_supported_file_types)

    path = _get_resource_path('kfp_test_operator.yaml')

    # Read contents of given path -- read_component_definition() returns a
    # a dictionary of component definition content indexed by path
    component_definition = reader.read_component_definition(path, {})[path]

    # Build entry for parsing
    entry = {
        "location_type": reader.resource_type,
        "location": path,
        "categories": ["Test"],
        "component_definition": component_definition
    }
    component_entry = SimpleNamespace(**entry)

    # Parse the component entry
    parser = KfpComponentParser()
    component = parser.parse(component_entry)[0]
    properties_json = ComponentRegistry.to_canvas_properties(component)

    # Ensure component parameters are prefixed (and system parameters are not) and all hold correct values
    assert properties_json['current_parameters']['label'] == ''
    assert properties_json['current_parameters']['component_source'] == component_entry.location
    assert properties_json['current_parameters']['elyra_test_string_no_default'] == ''
    assert properties_json['current_parameters']['elyra_test_string_default_value'] == 'default'
    assert properties_json['current_parameters']['elyra_test_string_default_empty'] == ''

    assert properties_json['current_parameters']['elyra_test_bool_default'] is False
    assert properties_json['current_parameters']['elyra_test_bool_false'] is False
    assert properties_json['current_parameters']['elyra_test_bool_true'] is True

    assert properties_json['current_parameters']['elyra_test_int_default'] == 0
    assert properties_json['current_parameters']['elyra_test_int_zero'] == 0
    assert properties_json['current_parameters']['elyra_test_int_non_zero'] == 1

    assert properties_json['current_parameters']['elyra_test_dict_default'] == ''  # {}
    assert properties_json['current_parameters']['elyra_test_list_default'] == ''  # []

    # Ensure that the 'required' attribute was set correctly. KFP components default to required
    # unless explicitly marked otherwise in component YAML.
    required_property = next(prop for prop in properties_json['uihints']['parameter_info']
                             if prop.get('parameter_ref') == 'elyra_test_required_property')
    assert required_property['data']['required'] is True

    optional_property = next(prop for prop in properties_json['uihints']['parameter_info']
                             if prop.get('parameter_ref') == 'elyra_test_optional_property')
    assert optional_property['data']['required'] is False

    default_required_property = next(prop for prop in properties_json['uihints']['parameter_info']
                                     if prop.get('parameter_ref') == 'elyra_test_required_property_default')
    assert default_required_property['data']['required'] is True

    # Ensure that type information is inferred correctly
    unusual_dict_property = next(prop for prop in properties_json['uihints']['parameter_info']
                                 if prop.get('parameter_ref') == 'elyra_test_unusual_type_dict')
    assert unusual_dict_property['data']['format'] == "dictionary"

    unusual_list_property = next(prop for prop in properties_json['uihints']['parameter_info']
                                 if prop.get('parameter_ref') == 'elyra_test_unusual_type_list')
    assert unusual_list_property['data']['format'] == "list"

    unusual_string_property = next(prop for prop in properties_json['uihints']['parameter_info']
                                   if prop.get('parameter_ref') == 'elyra_test_unusual_type_string')
    assert unusual_string_property['data']['format'] == "string"

    file_property = next(prop for prop in properties_json['uihints']['parameter_info']
                         if prop.get('parameter_ref') == 'elyra_test_unusual_type_file')
    assert file_property['data']['format'] == "inputpath"

    no_type_property = next(prop for prop in properties_json['uihints']['parameter_info']
                            if prop.get('parameter_ref') == 'elyra_test_unusual_type_notgiven')
    assert no_type_property['data']['format'] == "string"

    # Ensure descriptions are rendered properly with type hint in parentheses
    assert unusual_dict_property['description']['default'] == "The test command description " \
                                                              "(type: Dictionary of arrays)"
    assert unusual_list_property['description']['default'] == "The test command description (type: An array)"
    assert unusual_string_property['description']['default'] == "The test command description (type: A string)"
    assert file_property['description']['default'] == \
           "The test command description"  # No data type info is included in parentheses for inputPath variables
    assert no_type_property['description']['default'] == "The test command description (type: string)"


def test_parse_kfp_component_url():
    # Define the appropriate reader for a URL-type component definition
    kfp_supported_file_types = [".yaml"]
    reader = UrlComponentReader(kfp_supported_file_types)

    path = 'https://raw.githubusercontent.com/kubeflow/pipelines/1.4.1/components/notebooks/Run_notebook_using_papermill/component.yaml'  # noqa: E501

    # Read contents of given path -- read_component_definition() returns a
    # a dictionary of component definition content indexed by path
    component_definition = reader.read_component_definition(path, {})[path]

    # Build entry for parsing
    entry = {
        "location_type": reader.resource_type,
        "location": path,
        "categories": ["Test"],
        "component_definition": component_definition
    }
    component_entry = SimpleNamespace(**entry)

    # Parse the component entry
    parser = KfpComponentParser()
    component = parser.parse(component_entry)[0]
    properties_json = ComponentRegistry.to_canvas_properties(component)

    # Ensure component parameters are prefixed (and system parameters are not) and all hold correct values
    assert properties_json['current_parameters']['label'] == ''
    assert properties_json['current_parameters']['component_source'] == component_entry.location
    assert properties_json['current_parameters']['elyra_notebook'] == 'None'   # Default value for type `inputpath`
    assert properties_json['current_parameters']['elyra_parameters'] == '{}'
    assert properties_json['current_parameters']['elyra_packages_to_install'] == ''  # {}
    assert properties_json['current_parameters']['elyra_input_data'] == ''


def test_parse_kfp_component_file_no_inputs():
    # Define the appropriate reader for a filesystem-type component definition
    kfp_supported_file_types = [".yaml"]
    reader = FilesystemComponentReader(kfp_supported_file_types)

    path = _get_resource_path('kfp_test_operator_no_inputs.yaml')

    # Read contents of given path -- read_component_definition() returns a
    # a dictionary of component definition content indexed by path
    component_definition = reader.read_component_definition(path, {})[path]

    # Build entry for parsing
    entry = {
        "location_type": reader.resource_type,
        "location": path,
        "categories": ["Test"],
        "component_definition": component_definition
    }
    component_entry = SimpleNamespace(**entry)

    # Parse the component entry
    parser = KfpComponentParser()
    component = parser.parse(component_entry)[0]
    properties_json = ComponentRegistry.to_canvas_properties(component)

    # Properties JSON should only include the two parameters common to every
    # component:'label' and 'component_source', the component description if
    # exists (which it does for this component), and the output parameter for
    # this component
    num_common_params = 4
    assert len(properties_json['current_parameters'].keys()) == num_common_params
    assert len(properties_json['parameters']) == num_common_params
    assert len(properties_json['uihints']['parameter_info']) == num_common_params
    assert len(properties_json['uihints']['group_info'][0]['group_info']) == num_common_params

    # Ensure that template still renders the two common parameters correctly
    assert properties_json['current_parameters']['label'] == ""
    assert properties_json['current_parameters']['component_source'] == component_entry.location


async def test_parse_components_invalid_file():
    # Define the appropriate reader for a filesystem-type component definition
    kfp_supported_file_types = [".yaml"]
    reader = FilesystemComponentReader(kfp_supported_file_types)

    # Get path to an invalid component definition file and read contents
    path = _get_resource_path('kfp_test_operator_invalid.yaml')
    component_definition = reader.read_component_definition(path, {})
    assert component_definition == {}

    # Build entry for parsing
    entry = {
        "location_type": reader.resource_type,
        "location": path,
        "categories": ["Test"],
        "component_definition": component_definition
    }
    component_entry = SimpleNamespace(**entry)

    # Parse the component entry
    parser = KfpComponentParser()
    component = parser.parse(component_entry)
    assert component is None



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import ast
import re
from types import SimpleNamespace
from typing import Dict
from typing import List
from typing import Optional

from elyra.pipeline.component import Component
from elyra.pipeline.component import ComponentParameter
from elyra.pipeline.component import ComponentParser


class AirflowComponentParser(ComponentParser):
    _component_platform = "airflow"
    _file_types = [".py"]

    def get_catalog_entry_id_for_component(self, component_id: str) -> str:
        # Component ids are structure differently in Airflow to handle the case
        # where there are multiple classes in one operator file. The id queried
        # must be adjusted to match the id expected in the component_entry catalog.
        return component_id.split('_')[0]

    def parse(self, registry_entry: SimpleNamespace) -> Optional[List[Component]]:
        components: List[Component] = list()

        component_definition = registry_entry.component_definition
        if not component_definition:
            return None

        # Parse the component definition for all defined classes
        component_classes = self._get_all_classes(component_definition)
        for component_class in component_classes.keys():
            # Create a Component object for each class
            component_id = self.get_component_id(registry_entry.location, component_class)
            component_properties = self._parse_properties(component_definition, component_class)
            components.append(Component(id=component_id,
                                        name=component_class,
                                        description='',
                                        runtime=self.component_platform,
                                        location_type=registry_entry.location_type,
                                        location=registry_entry.location,
                                        properties=component_properties,
                                        categories=registry_entry.categories))

        return components

    def _get_all_classes(self, component_definition: str) -> Dict[str, Dict]:
        # Organize lines according to the class to which they belong
        classes = {}
        class_name = "no_class"
        classes["no_class"] = {"content": [], "args": []}
        class_regex = re.compile(r"class ([\w]+)\(\w*\):")
        for line in component_definition.split('\n'):
            # Remove any inline comments (must follow the '2 preceding spaces and one following space'
            # rule). This avoids the case where the default value of an __init__ arg contains '#'.
            line = re.sub(r"  # .*\n?", "", line)
            match = class_regex.search(line)
            if match:
                class_name = match.group(1)
                classes[class_name] = {"content": [], "args": []}
            classes[class_name]['content'].append(line)

        classes.pop("no_class")
        return classes

    def _get_class_with_classname(self, classname: str, component_definition: str) -> Dict[str, List]:
        classes = self._get_all_classes(component_definition)

        if classname not in classes.keys():
            raise ValueError(f"Component with class name {classname} not found")

        # Loop through classes to find init function for each class; grab init parameters as properties
        init_regex = re.compile(r"def __init__\(([\s\d\w,=\-\'\"\*\s\#.\\\/:?]*)\):")
        for class_name in classes:
            if class_name != classname:
                continue

            # Concatenate class body and search for __init__ function
            class_content = ''.join(classes[class_name]['content'])
            for match in init_regex.finditer(class_content):
                # Get list of parameter:default-value pairs
                classes[class_name]['args'] = [x.strip() for x in match.group(1).split(',')]

        return classes[classname]

    def _parse_properties(self, component_definition: str, component_class: str) -> List[ComponentParameter]:
        properties: List[ComponentParameter] = list()

        # NOTE: Currently no runtime-specific properties are needed, including runtime image. See
        # justification here: https://github.com/elyra-ai/elyra/issues/1912#issuecomment-879424452
        # properties.extend(self.get_runtime_specific_properties())

        # Retrieve the content of the specified class only
        component_definition = self._get_class_with_classname(component_class, component_definition)
        class_content = ''.join(component_definition.get('content'))
        for arg in component_definition.get('args'):
            # For each argument to the init function, build a new parameter and add to existing
            if arg in ['self', '*args', '**kwargs']:
                continue

            value = None
            if '=' in arg:
                arg, value = arg.split('=', 1)[:2]
                value = ast.literal_eval(value)
                if value and "\n" in str(value):
                    value = value.replace("\n", " ")

            # Search for :type [param] information in class docstring
            type_regex = re.compile(f":type {arg}:" + r"([\s\S]*?(?=:type|:param|\"\"\"|'''|\.\.))")
            match = type_regex.search(class_content)
            data_type = match.group(1).strip() if match else "string"

            # Search for :param [param] in class doctring to get description
            description = ""
            param_regex = re.compile(f":param {arg}:" + r"([\s\S]*?(?=:type|:param|\"\"\"|'''|\.\.))")
            match = param_regex.search(class_content)
            if match:
                description = match.group(1).strip().replace("\"", "'")

            # Amend description to include type information
            description = self._format_description(description=description, data_type=data_type)

            data_type_info = self.determine_type_information(data_type)
            if data_type_info.undetermined:
                self.log.warning(f"Data type from parsed data ('{data_type}') could not be determined. "
                                 f"Proceeding as if 'string' was detected.")

            properties.append(ComponentParameter(id=arg,
                                                 name=arg,
                                                 data_type=data_type_info.data_type,
                                                 value=(value or data_type_info.default_value),
                                                 description=description,
                                                 control_id=data_type_info.control_id))
        return properties

    def get_runtime_specific_properties(self) -> List[ComponentParameter]:
        """
        Define properties that are common to the Airflow runtime.
        """
        return [
            ComponentParameter(
                id="runtime_image",
                name="Runtime Image",
                data_type="string",
                value="",
                description="Container image used as execution environment.",
                control="custom",
                control_id="EnumControl",
                required=True,
            )
        ]



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""This module includes custom error classes pertaining to the metadata service."""


class MetadataNotFoundError(Exception):
    """Raised when a metadata instance is not found.

       Overrides FileNotFoundError to set contextual message text
       and includes the corresponding schemaspace.
    """
    def __init__(self, schemaspace: str, name: str):
        super().__init__("No such instance named '{}' was found in the {} schemaspace.".format(name, schemaspace))


class MetadataExistsError(Exception):
    """Raised when a metadata instance unexpectedly exists.

       Overrides FileExistsError to set contextual message text
       and includes the corresponding schemaspace.
    """
    def __init__(self, schemaspace: str, name: str):
        super().__init__("An instance named '{}' already exists in the {} schemaspace.".format(name, schemaspace))


class SchemaNotFoundError(Exception):
    """Raised when a schema instance is not found.

       Overrides FileNotFoundError to set contextual message text
       and includes the corresponding schemaspace.
    """
    def __init__(self, schemaspace: str, name: str):
        super().__init__("No such schema named '{}' was found in the {} schemaspace.".format(name, schemaspace))



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults


class TestOperatorNoInputs(BaseOperator):
    r"""
    Execute a test script.
    """

    @apply_defaults
    def __init__(self,
                 *args, **kwargs):

        super().__init__(*args, **kwargs)



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json
import os
from typing import Any
from typing import Dict
from typing import List
from typing import Optional


class AppDataBase:  # ABC
    """
    An abstraction for app_data based nodes
    """
    _node: Dict = None

    def __init__(self, node: Dict):
        """
        Constructor with the node json structure
        :param node: the node json
        """
        self._node = node

    @property
    def id(self) -> str:
        """
        The node id
        :return: the node unique identifier
        """
        return self._node.get('id')

    def get(self, key: str, default_value=None) -> Any:
        """
        Retrieve node values for a given key.
        These key/value pairs are stored in the app_data stanza
        :param key: The key to be retrieved
        :param default_value: a default value in case the key is not found
        :return: the value or the default_value if the key is not found
        """
        return self._node['app_data'].get(key, default_value)

    def set(self, key: str, value: Any):
        """
        Update node values for a given key.
        These key/value pairs are stored in the app_data stanza
        :param key: The key to be set
        :param value: The value to be set
        """
        if not key:
            raise ValueError("Key is required")

        if not value:
            raise ValueError("Value is required")

        self._node['app_data'][key] = value

    def to_dict(self) -> Dict:
        return self._node


class Pipeline(AppDataBase):
    _nodes: list = None

    def __init__(self, node: Dict):
        """
        The constructor with pipeline json structure
        :param node: the node pipeline
        """
        super().__init__(node)

    @property
    def version(self) -> int:
        """
        The pipelive version
        :return: The version
        """
        return int(self._node['app_data'].get('version'))

    @property
    def runtime(self) -> str:
        """
        The runtime associated with the pipeline
        :return: The runtime keyword
        """
        return self._node['app_data'].get('runtime')

    @property
    def runtime_config(self) -> str:
        """
        The runtime configuration associated with the pipeline
        :return: The runtime configuration key. This should be a valid key from the Runtimes metadata
        """
        return self._node['app_data'].get('runtime-config')

    @property
    def type(self):
        """
        The pipeline type
        :return: The runtime keyword associated with the pipeline or `generic`
        """
        type_description_to_type = {'Kubeflow Pipelines': 'kfp',
                                    'Apache Airflow': 'airflow',
                                    'Generic': 'generic'}

        if 'properties' in self._node['app_data']:
            pipeline_type_description = self._node['app_data']['properties'].get('runtime', 'Generic')
            if pipeline_type_description not in type_description_to_type.keys():
                raise ValueError(f'Unsupported pipeline runtime: {pipeline_type_description}')
            return type_description_to_type[pipeline_type_description]
        else:
            return type_description_to_type['Generic']

    @property
    def name(self) -> str:
        """
        The pipeline name
        :rtype: The pipeline name or `untitled`
        """
        return self._node['app_data'].get('name', self._node['app_data'].get('properties', {}).get('name', 'untitled'))

    @property
    def source(self) -> str:
        """
        The pipeline source
        :rtype: The pipeline source
        """
        return self._node['app_data'].get('source')

    @property
    def nodes(self) -> list:
        """
        The list of nodes for the pipeline
        :rtype: object
        """
        if 'nodes' not in self._node:
            raise ValueError("Pipeline is missing 'nodes' field.")

        if self._nodes is None:
            nodes: list = list()
            for node in self._node['nodes']:
                nodes.append(Node(node))

            self._nodes = nodes

        return self._nodes

    def get_property(self, key: str, default_value=None) -> Any:
        """
        Retrieve pipeline values for a given key.
        :param key: the key to be retrieved
        :param default_value: a default value in case the key is not found
        :return: the value or the default_value if the key is not found
        """
        return_value = default_value
        if 'properties' in self._node['app_data']:
            return_value = self._node['app_data']['properties'].get(key, default_value)

        return return_value

    def set_property(self, key: str, value: Any):
        """
        Update pipeline values for a given key.
        :param key: the key to be set
        :param value: the value to be set
        """
        if not key:
            raise ValueError("Key is required")

        if not value:
            raise ValueError("Value is required")

        self._node['app_data']['properties'][key] = value


class Node(AppDataBase):
    def __init__(self, node: Dict):
        super().__init__(node)

    @property
    def type(self) -> str:
        """
        The node type
        :return: type (e.g. execution_node, super_node)
        """
        return self._node.get('type')

    @property
    def op(self) -> str:
        """
        The node op, which identify the operation to be executed
        :return: op (e.g. execute-notebook-node)
        """
        return self._node.get('op')

    @property
    def label(self) -> str:
        """
        The node label
        :return:  node label
        """
        return self._node['app_data']['ui_data'].get('label', self._node['app_data'].get('label', None))

    @property
    def subflow_pipeline_id(self) -> Pipeline:
        """
        The Super Node pipeline reference. Only available when type is a super node.
        :return:
        """
        if self._node['type'] != 'super_node':
            raise ValueError("Node must be a super_node in order to retrieve a subflow pipeline id")

        if 'subflow_ref' in self._node:
            return self._node['subflow_ref'].get('pipeline_id_ref')
        else:
            return None

    @property
    def component_links(self) -> List:
        """
        Retrieve component links to other components.
        :return: the list of links associated with this node or an empty list if none are found
        """
        if self.type in ['execution_node', 'super_node']:
            return self._node['inputs'][0].get('links', [])
        else:
            #  binding nodes do not contain links
            return []

    def get_component_parameter(self, key: str, default_value=None) -> Any:
        """
        Retrieve component parameter values.
        These key/value pairs are stored in app_data.component_parameters
        :param key: the parameter key to be retrieved
        :param default_value: a default value in case the key is not found
        :return: the value or the default value if the key is not found
        """
        return self._node['app_data']['component_parameters'].get(key, default_value)

    def set_component_parameter(self, key: str, value: Any):
        """
        Update component parameter values for a given key.
        These key/value pairs are stored in app_data.component_parameters
        :param key: The parameter key to be retrieved
        :param value: the value to be set
        """
        if not key:
            raise ValueError("Key is required")

        if not value:
            raise ValueError("Value is required")

        self._node['app_data']['component_parameters'][key] = value


class PipelineDefinition(object):
    """
    Represents a helper class to manipulate pipeline json structure
    """
    _pipelines: list = None
    _primary_pipeline: Pipeline = None
    _validated: bool = False
    _validation_issues: list = None

    def __init__(self, pipeline_path: Optional[str] = None,
                 pipeline_definition: Optional[Dict] = None,
                 validate: bool = False):
        """
        The constructor enables either passing a pipeline path or the content of the pipeline definition.
        :param pipeline_path: this is the path to a pipeline
        :param pipeline_definition: this is the piepline json
        :param validate: flag to turn validation during pipeline initialization
        """
        if not pipeline_path and not pipeline_definition:
            # at least one parameter should be provided
            raise ValueError("At least one parameter must be provided ('pipeline_path' or 'pipeline_definition')")
        if pipeline_path and pipeline_definition:
            # only one parameter should be provided
            raise ValueError("Only one parameter should be provided ('pipeline_path' or 'pipeline_definition')")

        if pipeline_path:
            # supporting loading pipeline from file
            if not os.path.exists(pipeline_path):
                raise ValueError(f"Pipeline file not found: '{pipeline_path}'\n")

            with open(pipeline_path) as f:
                try:
                    self._pipeline_definition = json.load(f)
                except ValueError as ve:
                    raise ValueError(f"Pipeline file is invalid: \n {ve}")
        else:
            # supporting passing the pipeline definition directly
            self._pipeline_definition = pipeline_definition

        if validate:
            self.validate()

    @property
    def id(self) -> str:
        """
        The pipeline definition id
        :return: the unid
        """
        return self._pipeline_definition.get('id')

    @property
    def schema_version(self) -> str:
        """
        The schema used by the Pipeline definition
        :return: the version
        """
        return self._pipeline_definition.get('version')

    @property
    def pipelines(self) -> list:
        """
        The list of pipelines defined in the pipeline definition
        :return: the list of pipelines
        """
        if not self._pipelines:
            if 'pipelines' not in self._pipeline_definition:
                raise ValueError("Pipeline is missing 'pipelines' field.")
            elif len(self._pipeline_definition['pipelines']) == 0:
                raise ValueError("Pipeline has zero length 'pipelines' field.")

            pipelines: list = list()
            for pipeline in self._pipeline_definition['pipelines']:
                pipelines.append(Pipeline(pipeline))

            self._pipelines = pipelines

        return self._pipelines

    @property
    def primary_pipeline(self) -> Pipeline:
        """
        The primary pipeline associated with this pipeline definition
        :return: the primary pipeline
        """
        if not self._primary_pipeline:
            if "pipelines" not in self._pipeline_definition:
                raise ValueError("Pipeline is missing 'pipelines' field.")
            elif len(self._pipeline_definition['pipelines']) == 0:
                raise ValueError("Pipeline has zero length 'pipelines' field.")

            # Find primary pipeline
            self._primary_pipeline = self.get_pipeline_definition(self._pipeline_definition.get('primary_pipeline'))

            assert self._primary_pipeline is not None, "No primary pipeline was found"

        return self._primary_pipeline

    def validate(self) -> list:
        """
        Validates the pipeline definition structure and semantics
        :return: the list of issues found
        """
        # If it has been validated before
        if self._validated:
            # return current list of issues
            return self._validation_issues

        # Has not been validated before
        validation_issues = list()
        # Validate pipeline schema version
        if 'version' not in self._pipeline_definition:
            validation_issues.append("Pipeline schema version field is missing.")
        elif not isinstance(self._pipeline_definition['version'], str):
            validation_issues.append("Pipeline schema version field should be a string.")

        # Validate pipelines
        if 'pipelines' not in self._pipeline_definition:
            validation_issues.append("Pipeline is missing 'pipelines' field.")
        elif not isinstance(self._pipeline_definition["pipelines"], list):
            validation_issues.append("Field 'pipelines' should be a list.")
        elif len(self._pipeline_definition['pipelines']) == 0:
            validation_issues.append("Pipeline has zero length 'pipelines' field.")

        # Validate primary pipeline
        if 'primary_pipeline' not in self._pipeline_definition:
            validation_issues.append("Could not determine the primary pipeline.")
        elif not isinstance(self._pipeline_definition["primary_pipeline"], str):
            validation_issues.append("Field 'primary_pipeline' should be a string.")

        primary_pipeline = self.get_pipeline_definition(self._pipeline_definition.get('primary_pipeline'))
        if not primary_pipeline:
            validation_issues.append("No primary pipeline was found")
        else:
            primary_pipeline = primary_pipeline.to_dict()
            # Validate primary pipeline structure
            if 'app_data' not in primary_pipeline:
                validation_issues.append("Primary pipeline is missing the 'app_data' field.")
            else:
                if 'version' not in primary_pipeline['app_data']:
                    validation_issues.append("Primary pipeline is missing the 'version' field.")
                if 'properties' not in primary_pipeline['app_data']:
                    validation_issues.append("Node is missing 'properties' field.")
                elif len(primary_pipeline['app_data']['properties']) == 0:
                    validation_issues.append("Pipeline has zero length 'properties' field.")

            if 'nodes' not in primary_pipeline or len(primary_pipeline['nodes']) == 0:
                validation_issues.append("At least one node must exist in the primary pipeline.")
            else:
                for node in primary_pipeline['nodes']:
                    if "component_parameters" not in node['app_data']:
                        validation_issues.append("Node is missing 'component_parameters' field")

        return validation_issues

    def is_valid(self) -> bool:
        """
        Represents whether or not the pipeline structure is valid
        :return: True for a valid pipeline definition
        """
        return len(self.validate()) == 0

    def to_dict(self) -> Dict:
        """
        The raw contents of the pipeline definition json
        :rtype: object
        """
        return self._pipeline_definition

    def get_pipeline_definition(self, pipeline_id) -> Any:
        """
        Retrieve a given pipeline from the pipeline definition
        :param pipeline_id: the pipeline unique identifier
        :return: the pipeline or None
        """
        if 'pipelines' in self._pipeline_definition:
            for pipeline in self._pipeline_definition["pipelines"]:
                if pipeline['id'] == pipeline_id:
                    return Pipeline(pipeline)

        return None

    def get_node(self, node_id: str):
        """
        Given a node id returns the associated node object in the pipeline
        :param node_id: the node id
        :return: the node object or None
        """
        for pipeline in self._pipelines:
            for node in pipeline.nodes:
                if node.id == node_id:
                    return node
        return None

    def get_supernodes(self) -> List[Node]:
        """
        Returns a list of all supernodes in the pipeline
        :return:
        """
        supernode_list = []
        for pipeline in self._pipelines:
            for node in pipeline.nodes:
                if node.type == "super_node":
                    supernode_list.append(node)
        return supernode_list



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys

import pytest

from elyra.pipeline.pipeline import GenericOperation
from elyra.pipeline.pipeline import Operation
from elyra.pipeline.pipeline import Pipeline


@pytest.fixture
def good_operation():
    component_parameters = {
        'filename': 'elyra/pipeline/tests/resources/archive/test.ipynb',
        'runtime_image': 'tensorflow/tensorflow:latest'
    }
    test_operation = GenericOperation(id='test-id',
                                      type='execution-node',
                                      classifier='execute-notebook-node',
                                      name='test',
                                      component_params=component_parameters)
    return test_operation


@pytest.fixture
def good_pipeline():
    test_pipeline = Pipeline(id='Random-UUID-123123123123123',
                             name='test-pipeline',
                             runtime='kfp',
                             runtime_config='default_kfp')
    return test_pipeline


def test_create_operation_minimal(good_operation):
    test_operation = good_operation

    assert test_operation.id == 'test-id'
    assert test_operation.type == 'execution-node'
    assert test_operation.classifier == 'execute-notebook-node'
    assert test_operation.name == 'test'
    assert test_operation.filename == 'elyra/pipeline/tests/resources/archive/test.ipynb'
    assert test_operation.runtime_image == 'tensorflow/tensorflow:latest'
    assert test_operation.name == 'test'


def test_create_operation_with_dependencies():
    dependencies = ['elyra/pipline/tests/resources', 'elyra/pipline/tests/resources/archive']

    component_parameters = {
        'filename': 'elyra/pipeline/tests/resources/archive/test.ipynb',
        'dependencies': dependencies,
        'runtime_image': 'tensorflow/tensorflow:latest'
    }
    test_operation = GenericOperation(id='test-id',
                                      type='execution-node',
                                      classifier='execute-notebook-node',
                                      name='test',
                                      component_params=component_parameters)

    assert test_operation.dependencies == dependencies


def test_create_operation_include_subdirectories():
    include_subdirectories = True

    component_parameters = {
        'filename': 'elyra/pipeline/tests/resources/archive/test.ipynb',
        'include_subdirectories': include_subdirectories,
        'runtime_image': 'tensorflow/tensorflow:latest'
    }
    test_operation = GenericOperation(id='test-id',
                                      type='execution-node',
                                      classifier='execute-notebook-node',
                                      name='test',
                                      component_params=component_parameters)

    assert test_operation.include_subdirectories == include_subdirectories


def test_create_operation_with_environmental_variables():
    env_variables = ['FOO="Bar"', 'BAR="Foo']

    component_parameters = {
        'filename': 'elyra/pipeline/tests/resources/archive/test.ipynb',
        'env_vars': env_variables,
        'runtime_image': 'tensorflow/tensorflow:latest'
    }
    test_operation = GenericOperation(id='test-id',
                                      type='execution-node',
                                      classifier='execute-notebook-node',
                                      name='test',
                                      component_params=component_parameters)

    assert test_operation.env_vars == env_variables


def test_create_operation_with_inputs():
    inputs = ["input1.txt", "input2.txt"]

    component_parameters = {
        'filename': 'elyra/pipeline/tests/resources/archive/test.ipynb',
        'inputs': inputs,
        'runtime_image': 'tensorflow/tensorflow:latest'
    }
    test_operation = GenericOperation(id='test-id',
                                      type='execution-node',
                                      classifier='execute-notebook-node',
                                      name='test',
                                      component_params=component_parameters)

    assert test_operation.inputs == inputs


def test_create_operation_with_outputs():
    outputs = ["output1.txt", "output2.txt"]

    component_parameters = {
        'filename': 'elyra/pipeline/tests/resources/archive/test.ipynb',
        'outputs': outputs,
        'runtime_image': 'tensorflow/tensorflow:latest'
    }
    test_operation = GenericOperation(id='test-id',
                                      type='execution-node',
                                      classifier='execute-notebook-node',
                                      name='test',
                                      component_params=component_parameters)

    assert test_operation.outputs == outputs


def test_create_operation_with_parent_operations():
    parent_operation_ids = ['id-123123-123123-123123', 'id-456456-456456-456456']

    component_parameters = {
        'filename': 'elyra/pipeline/tests/resources/archive/test.ipynb',
        'runtime_image': 'tensorflow/tensorflow:latest'
    }
    test_operation = GenericOperation(id='test-id',
                                      type='execution-node',
                                      classifier='execute-notebook-node',
                                      name='test',
                                      parent_operation_ids=parent_operation_ids,
                                      component_params=component_parameters)

    assert test_operation.parent_operation_ids == parent_operation_ids


def test_create_operation_correct_naming():
    label = 'test.ipynb'
    filename = 'elyra/pipeline/tests/resources/archive/' + label

    component_parameters = {
        'filename': filename,
        'runtime_image': 'tensorflow/tensorflow:latest'
    }
    test_operation = GenericOperation(id='test-id',
                                      type='execution-node',
                                      classifier='execute-notebook-node',
                                      name=label,
                                      component_params=component_parameters)

    assert test_operation.name == label.split('.')[0]


def test_fail_create_operation_missing_id():
    component_parameters = {
        'filename': 'elyra/pipeline/tests/resources/archive/test.ipynb',
        'runtime_image': 'tensorflow/tensorflow:latest'
    }
    with pytest.raises(TypeError):
        Operation(type='execution-node',
                  classifier='execute-notebook-node',
                  name='test',
                  component_params=component_parameters)


def test_fail_create_operation_missing_type():
    component_parameters = {
        'filename': 'elyra/pipeline/tests/resources/archive/test.ipynb',
        'runtime_image': 'tensorflow/tensorflow:latest'
    }
    with pytest.raises(TypeError):
        GenericOperation(id='test-id',
                         classifier='execute-notebook-node',
                         name='test',
                         component_params=component_parameters)


def test_fail_create_operation_missing_classifier():
    component_parameters = {
        'filename': 'elyra/pipeline/tests/resources/archive/test.ipynb',
        'runtime_image': 'tensorflow/tensorflow:latest'
    }
    with pytest.raises(TypeError):
        Operation(id='test-id',
                  type='execution-node',
                  name='test',
                  component_params=component_parameters)


def test_fail_create_operation_missing_runtime_image():
    component_parameters = {
        'filename': 'elyra/pipeline/tests/resources/archive/test.ipynb'
    }
    with pytest.raises(ValueError):
        GenericOperation(id='test-id',
                         type='execution-node',
                         classifier='execute-notebook-node',
                         name='test',
                         component_params=component_parameters)


def test_fail_create_operation_missing_name():
    component_parameters = {
        'filename': 'elyra/pipeline/tests/resources/archive/test.ipynb',
        'runtime_image': 'tensorflow/tensorflow:latest'
    }
    with pytest.raises(TypeError):
        GenericOperation(id='test-id',
                         type='execution-node',
                         classifier='execute-notebook-node',
                         component_params=component_parameters)


def test_fail_operations_are_equal(good_operation):
    parent_operation_ids = ['id-123123-123123-123123', 'id-456456-456456-456456']
    component_parameters = {
        'filename': 'elyra/pipeline/tests/resources/archive/test.ipynb',
        'runtime_image': 'tensorflow/tensorflow:latest'
    }
    compare_operation = GenericOperation(id='test-id',
                                         type='execution-node',
                                         classifier='execute-notebook-node',
                                         name='test',
                                         parent_operation_ids=parent_operation_ids,
                                         component_params=component_parameters)
    with pytest.raises(AssertionError):
        assert compare_operation == good_operation


def test_operations_are_equal(good_operation):
    component_parameters = {
        'filename': 'elyra/pipeline/tests/resources/archive/test.ipynb',
        'runtime_image': 'tensorflow/tensorflow:latest'
    }
    compare_operation = GenericOperation(id='test-id',
                                         type='execution-node',
                                         classifier='execute-notebook-node',
                                         name='test',
                                         component_params=component_parameters)

    assert compare_operation == good_operation


def test_create_pipeline_minimal(good_pipeline):
    test_pipeline = good_pipeline

    assert test_pipeline.id == 'Random-UUID-123123123123123'
    assert test_pipeline.name == 'test-pipeline'
    assert test_pipeline.runtime == 'kfp'
    assert test_pipeline.runtime_config == 'default_kfp'


def test_fail_create_pipeline_missing_id():
    with pytest.raises(TypeError):
        Pipeline(name='test-pipeline',
                 runtime='kfp',
                 runtime_config='default_kfp')


def test_fail_create_pipeline_missing_name():
    with pytest.raises(TypeError):
        Pipeline(id='Random-UUID-123123123123123',
                 runtime='kfp',
                 runtime_config='default_kfp')


def test_fail_create_pipeline_missing_runtime():
    with pytest.raises(TypeError):
        Pipeline(id='Random-UUID-123123123123123',
                 name='test-pipeline',
                 runtime_config='default_kfp')


def test_fail_create_pipeline_missing_runtime_config():
    with pytest.raises(TypeError):
        Pipeline(id='Random-UUID-123123123123123',
                 name='test-pipeline',
                 runtime='kfp')


def test_pipelines_are_equal(good_pipeline):
    compare_pipeline = Pipeline(id='Random-UUID-123123123123123',
                                name='test-pipeline',
                                runtime='kfp',
                                runtime_config='default_kfp')

    assert compare_pipeline == good_pipeline


def test_fail_pipelines_are_equal(good_pipeline):
    test_operations_dict = {'123123123': good_operation,
                            '234234234': good_operation}

    compare_pipeline = Pipeline(id='Random-UUID-123123123123123',
                                name='test-pipeline',
                                runtime='kfp',
                                runtime_config='default_kfp')

    for key, operation in test_operations_dict.items():
        compare_pipeline.operations[key] = operation

    with pytest.raises(AssertionError):
        assert compare_pipeline == good_pipeline


def test_env_list_to_dict_function():
    env_variables = ['KEY=value', None, '', '  =empty_key', '=no_key', 'EMPTY_VALUE=  ',
                     'NO_VALUE=', 'KEY2=value2', 'TWO_EQUALS=KEY=value', '==']
    env_variables_dict = {"KEY": "value", "KEY2": "value2", "EMPTY_VALUE": "  ", "TWO_EQUALS": "KEY=value"}

    component_parameters = {
        'filename': 'elyra/pipeline/tests/resources/archive/test.ipynb',
        'env_vars': env_variables,
        'runtime_image': 'tensorflow/tensorflow:latest'
    }
    test_operation = GenericOperation(id='test-id',
                                      type='execution-node',
                                      classifier='execute-notebook-node',
                                      name='test',
                                      component_params=component_parameters)

    assert test_operation.env_vars_as_dict() == env_variables_dict


def test_validate_resource_values():

    component_parameters = {
        'filename': 'elyra/pipeline/tests/resources/archive/test.ipynb',
        'cpu': '4',
        'gpu': '6',
        'memory': '10',
        'runtime_image': 'tensorflow/tensorflow:latest'
    }
    test_operation = GenericOperation(id='test-id',
                                      type='execution-node',
                                      classifier='execute-notebook-node',
                                      name='test',
                                      component_params=component_parameters)

    assert test_operation.cpu == '4'
    assert test_operation.gpu == '6'
    assert test_operation.memory == '10'


def test_validate_resource_values_as_none():

    component_parameters = {
        'filename': 'elyra/pipeline/tests/resources/archive/test.ipynb',
        'runtime_image': 'tensorflow/tensorflow:latest'
    }
    test_operation = GenericOperation(id='test-id',
                                      type='execution-node',
                                      classifier='execute-notebook-node',
                                      name='test',
                                      component_params=component_parameters)

    assert test_operation.cpu is None
    assert test_operation.gpu is None
    assert test_operation.memory is None


def test_validate_gpu_accepts_zero_as_value():

    component_parameters = {
        'filename': 'elyra/pipeline/tests/resources/archive/test.ipynb',
        'cpu': '4',
        'gpu': '0',
        'memory': '10',
        'runtime_image': 'tensorflow/tensorflow:latest'
    }
    test_operation = GenericOperation(id='test-id',
                                      type='execution-node',
                                      classifier='execute-notebook-node',
                                      name='test',
                                      component_params=component_parameters)

    assert test_operation.gpu == '0'


def test_validate_max_resource_value():
    system_max_size = str(sys.maxsize - 1)

    component_parameters = {
        'filename': 'elyra/pipeline/tests/resources/archive/test.ipynb',
        'memory': system_max_size,
        'runtime_image': 'tensorflow/tensorflow:latest'
    }
    test_operation = GenericOperation(id='test-id',
                                      type='execution-node',
                                      classifier='execute-notebook-node',
                                      name='test',
                                      component_params=component_parameters)

    assert test_operation.memory == system_max_size


def test_fail_validate_max_resource_value_exceeded():
    system_max_size = str(sys.maxsize)

    component_parameters = {
        'filename': 'elyra/pipeline/tests/resources/archive/test.ipynb',
        'memory': system_max_size,
        'runtime_image': 'tensorflow/tensorflow:latest'
    }
    with pytest.raises(ValueError):
        GenericOperation(id='test-id',
                         type='execution-node',
                         classifier='execute-notebook-node',
                         name='test',
                         component_params=component_parameters)


def test_fail_creating_operation_with_negative_gpu_resources():
    component_parameters = {
        'filename': 'elyra/pipeline/tests/resources/archive/test.ipynb',
        'gpu': '-1',
        'runtime_image': 'tensorflow/tensorflow:latest'
    }
    with pytest.raises(ValueError):
        GenericOperation(id='test-id',
                         type='execution-node',
                         classifier='execute-notebook-node',
                         name='test',
                         component_params=component_parameters)


def test_fail_creating_operation_with_0_cpu_resources():
    component_parameters = {
        'filename': 'elyra/pipeline/tests/resources/archive/test.ipynb',
        'cpu': '0',
        'runtime_image': 'tensorflow/tensorflow:latest'
    }
    with pytest.raises(ValueError):
        GenericOperation(id='test-id',
                         type='execution-node',
                         classifier='execute-notebook-node',
                         name='test',
                         component_params=component_parameters)


def test_fail_creating_operation_with_negative_cpu_resources():
    component_parameters = {
        'filename': 'elyra/pipeline/tests/resources/archive/test.ipynb',
        'cpu': '-1',
        'runtime_image': 'tensorflow/tensorflow:latest'
    }
    with pytest.raises(ValueError):
        GenericOperation(id='test-id',
                         type='execution-node',
                         classifier='execute-notebook-node',
                         name='test',
                         component_params=component_parameters)


def test_fail_creating_operation_with_0_memory_resources():
    component_parameters = {
        'filename': 'elyra/pipeline/tests/resources/archive/test.ipynb',
        'memory': '0',
        'runtime_image': 'tensorflow/tensorflow:latest'
    }
    with pytest.raises(ValueError):
        GenericOperation(id='test-id',
                         type='execution-node',
                         classifier='execute-notebook-node',
                         name='test',
                         component_params=component_parameters)


def test_fail_creating_operation_with_negative_memory_resources():
    component_parameters = {
        'filename': 'elyra/pipeline/tests/resources/archive/test.ipynb',
        'memory': '-1',
        'runtime_image': 'tensorflow/tensorflow:latest'
    }
    with pytest.raises(ValueError):
        GenericOperation(id='test-id',
                         type='execution-node',
                         classifier='execute-notebook-node',
                         name='test',
                         component_params=component_parameters)


def test_scrub_list_function():
    env_variables_input = ['FOO=Bar', 'BAR=Foo', None, '']
    env_variables_output = ['FOO=Bar', 'BAR=Foo']

    assert Operation._scrub_list(env_variables_input) == env_variables_output



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json

import pytest

from elyra.tests.contents.test_utils import create_dir
from elyra.tests.contents.test_utils import create_file
from elyra.tests.contents.test_utils import empty_notebook_content
from elyra.tests.contents.test_utils import notebook_content
from elyra.tests.contents.test_utils import python_content
from elyra.tests.contents.test_utils import r_content
from elyra.tests.contents.test_utils import text_content


@pytest.fixture
def directory_name():
    return "dir.py"


@pytest.fixture
def text_filename():
    return "test.txt"


@pytest.fixture
def notebook_filename():
    return "test.ipynb"


@pytest.fixture
def python_filename():
    return "test.py"


@pytest.fixture
def r_filename():
    return "test.r"


@pytest.fixture
def create_directory(jp_root_dir, directory_name):
    create_dir(jp_root_dir, directory_name)


@pytest.fixture
def create_text_file(jp_root_dir, text_filename):
    create_file(jp_root_dir, text_filename, text_content)


@pytest.fixture
def create_notebook_file(jp_root_dir, notebook_filename):
    create_file(jp_root_dir, notebook_filename, json.dumps(notebook_content))


@pytest.fixture
def create_python_file(jp_root_dir, python_filename):
    create_file(jp_root_dir, python_filename, python_content)


@pytest.fixture
def create_r_file(jp_root_dir, r_filename):
    create_file(jp_root_dir, r_filename, r_content)


@pytest.fixture
def create_empty_notebook_file(jp_root_dir, notebook_filename):
    create_file(jp_root_dir, notebook_filename, json.dumps(empty_notebook_content))


# Set Elyra server extension as enabled (overriding server_config fixture from jupyter_server)
@pytest.fixture
def jp_server_config():
    return {
        "ServerApp": {
            "jpserver_extensions": {
                "elyra": True
            }
        }
    }



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import errno
import os


def create_dir(location, dir_name):
    try:
        dir_path = os.path.join(location, dir_name)
        os.mkdir(dir_path)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise


def create_file(location, file_name, content):
    try:
        os.makedirs(location)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

    resource = os.path.join(location, file_name)
    with open(resource, 'w', encoding='utf-8') as f:
        f.write(content)


expected_response = {
    "env_vars": {
        "VAR1": "newvalue",
        "VAR2": None,
        "VAR3": None,
        "VAR4": None,
        "VAR5": "localhost",
        "VAR6": "6",
        "VAR7": "value7",
        "VAR8": None
    },
    "inputs": [],
    "outputs": []
}

expected_response_empty = {
    "env_vars": {},
    "inputs": [],
    "outputs": []
}

text_content = "This is a text file."

notebook_content = {
    "cells": [
        {
            "cell_type": "markdown",
            "id": "advanced-touch",
            "metadata": {},
            "source": [
                "# Python Notebook with Environment Variables\n",
                "\n",
                "This python Notebook contains various environment variables to test the parser functionality.",
            ],
        },
        {
            "cell_type": "code",
            "execution_count": 0,
            "id": "regional-indie",
            "metadata": {},
            "outputs": [],
            "source": [
                "import os\n",
                "\n",
                'os.getenv("VAR1")\n',
                'os.environ["VAR2"]\n',
                'os.environ.get("VAR3")\n',
                "\n",
                "print(os.environ['VAR4'])\n",
                "print(os.getenv(\"VAR5\", 'localhost'))",
            ],
        },
        {
            "cell_type": "code",
            "execution_count": 1,
            "id": "completed-timothy",
            "metadata": {},
            "outputs": [],
            "source": [
                "os.environ['VAR6'] = \"6\"\n",
                "print(os.environ.get('VAR7', 'value7'))\n",
                "os.getenv('VAR8')\n",
                "\n",
                'os.environ["VAR1"] = "newvalue"',
            ],
        },
    ],
    "metadata": {
        "kernelspec": {
            "display_name": "Python 3",
            "language": "python",
            "name": "python3",
        },
        "language_info": {
            "codemirror_mode": {"name": "ipython", "version": 3},
            "file_extension": ".py",
            "mimetype": "text/x-python",
            "name": "python",
            "nbconvert_exporter": "python",
            "pygments_lexer": "ipython3",
            "version": "3.9.1",
        },
    },
    "nbformat": 4,
    "nbformat_minor": 5,
}

python_content = '''
    import os

    os.getenv("VAR1")
    os.environ["VAR2"]
    os.environ.get("VAR3")

    print(os.environ['VAR4'])
    print(os.getenv("VAR5", 'localhost'))

    os.environ['VAR6'] = "6"
    print(os.environ.get('VAR7', 'value7'))
    os.getenv('VAR8')

    os.environ["VAR1"] = "newvalue"
'''

r_content = '''
    Sys.setenv(VAR1 = "newvalue")
    Sys.getenv(VAR2)

    Sys.getenv("VAR3")
    Sys.getenv('VAR4')

    Sys.setenv('VAR5' = 'localhost')
    Sys.setenv("VAR6" = 6)

     Sys.setenv(VAR7 = "value7")
    Sys.getenv('VAR8')
'''

empty_notebook_content = {
    "cells": [
        {
            "cell_type": "markdown",
            "id": "literary-parts",
            "metadata": {},
            "source": [
                "# Python Notebook with No Environment Variables\n",
                "\n",
                "This python Notebook contains no environment variables to test the parser functionality.",
            ],
        },
        {
            "cell_type": "code",
            "execution_count": 0,
            "id": "dental-manchester",
            "metadata": {},
            "outputs": [],
            "source": ["import os\n", "\n", "print(os.cwd())"],
        },
    ],
    "metadata": {
        "kernelspec": {
            "display_name": "Python 3",
            "language": "python",
            "name": "python3",
        },
        "language_info": {
            "codemirror_mode": {"name": "ipython", "version": 3},
            "file_extension": ".py",
            "mimetype": "text/x-python",
            "name": "python",
            "nbconvert_exporter": "python",
            "pygments_lexer": "ipython3",
            "version": "3.9.1",
        },
    },
    "nbformat": 4,
    "nbformat_minor": 5,
}



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import string

import pytest

from elyra.kfp.operator import ExecuteFileOp


def test_fail_without_cos_endpoint():
    with pytest.raises(TypeError):
        ExecuteFileOp(name="test",
                      pipeline_name="test-pipeline",
                      experiment_name="experiment-name",
                      notebook="test_notebook.ipynb",
                      cos_bucket="test_bucket",
                      cos_directory="test_directory",
                      cos_dependencies_archive="test_archive.tgz",
                      image="test/image:dev")


def test_fail_without_cos_bucket():
    with pytest.raises(TypeError):
        ExecuteFileOp(name="test",
                      pipeline_name="test-pipeline",
                      experiment_name="experiment-name",
                      notebook="test_notebook.ipynb",
                      cos_endpoint="http://testserver:32525",
                      cos_directory="test_directory",
                      cos_dependencies_archive="test_archive.tgz",
                      image="test/image:dev")


def test_fail_without_cos_directory():
    with pytest.raises(TypeError):
        ExecuteFileOp(name="test",
                      pipeline_name="test-pipeline",
                      experiment_name="experiment-name",
                      notebook="test_notebook.ipynb",
                      cos_endpoint="http://testserver:32525",
                      cos_bucket="test_bucket",
                      cos_dependencies_archive="test_archive.tgz",
                      image="test/image:dev")


def test_fail_without_cos_dependencies_archive():
    with pytest.raises(TypeError):
        ExecuteFileOp(name="test",
                      pipeline_name="test-pipeline",
                      experiment_name="experiment-name",
                      notebook="test_notebook.ipynb",
                      cos_endpoint="http://testserver:32525",
                      cos_bucket="test_bucket",
                      cos_directory="test_directory",
                      image="test/image:dev")


def test_fail_without_runtime_image():
    with pytest.raises(ValueError) as error_info:
        ExecuteFileOp(name="test",
                      pipeline_name="test-pipeline",
                      experiment_name="experiment-name",
                      notebook="test_notebook.ipynb",
                      cos_endpoint="http://testserver:32525",
                      cos_bucket="test_bucket",
                      cos_directory="test_directory",
                      cos_dependencies_archive="test_archive.tgz")
    assert "You need to provide an image." == str(error_info.value)


def test_fail_without_notebook():
    with pytest.raises(TypeError):
        ExecuteFileOp(name="test",
                      pipeline_name="test-pipeline",
                      experiment_name="experiment-name",
                      cos_endpoint="http://testserver:32525",
                      cos_bucket="test_bucket",
                      cos_directory="test_directory",
                      cos_dependencies_archive="test_archive.tgz",
                      image="test/image:dev")


def test_fail_without_name():
    with pytest.raises(TypeError):
        ExecuteFileOp(pipeline_name="test-pipeline",
                      experiment_name="experiment-name",
                      notebook="test_notebook.ipynb",
                      cos_endpoint="http://testserver:32525",
                      cos_bucket="test_bucket",
                      cos_directory="test_directory",
                      cos_dependencies_archive="test_archive.tgz",
                      image="test/image:dev")


def test_fail_with_empty_string_as_name():
    with pytest.raises(ValueError):
        ExecuteFileOp(name="",
                      pipeline_name="test-pipeline",
                      experiment_name="experiment-name",
                      notebook="test_notebook.ipynb",
                      cos_endpoint="http://testserver:32525",
                      cos_bucket="test_bucket",
                      cos_directory="test_directory",
                      cos_dependencies_archive="test_archive.tgz",
                      image="test/image:dev")


def test_fail_with_empty_string_as_notebook():
    with pytest.raises(ValueError) as error_info:
        ExecuteFileOp(name="test",
                      pipeline_name="test-pipeline",
                      experiment_name="experiment-name",
                      notebook="",
                      cos_endpoint="http://testserver:32525",
                      cos_bucket="test_bucket",
                      cos_directory="test_directory",
                      cos_dependencies_archive="test_archive.tgz",
                      image="test/image:dev")
    assert "You need to provide a notebook." == str(error_info.value)


def test_fail_without_pipeline_name():
    with pytest.raises(TypeError):
        ExecuteFileOp(name="test",
                      experiment_name="experiment-name",
                      notebook="test_notebook.ipynb",
                      cos_endpoint="http://testserver:32525",
                      cos_bucket="test_bucket",
                      cos_directory="test_directory",
                      cos_dependencies_archive="test_archive.tgz",
                      image="test/image:dev")


def test_fail_without_experiment_name():
    with pytest.raises(TypeError):
        ExecuteFileOp(name="test",
                      pipeline_name="test-pipeline",
                      notebook="test_notebook.ipynb",
                      cos_endpoint="http://testserver:32525",
                      cos_bucket="test_bucket",
                      cos_directory="test_directory",
                      cos_dependencies_archive="test_archive.tgz",
                      image="test/image:dev")


def test_properly_set_notebook_name_when_in_subdirectory():
    notebook_op = ExecuteFileOp(name="test",
                                pipeline_name="test-pipeline",
                                experiment_name="experiment-name",
                                notebook="foo/test_notebook.ipynb",
                                cos_endpoint="http://testserver:32525",
                                cos_bucket="test_bucket",
                                cos_directory="test_directory",
                                cos_dependencies_archive="test_archive.tgz",
                                image="test/image:dev")
    assert "test_notebook.ipynb" == notebook_op.notebook_name


def test_properly_set_python_script_name_when_in_subdirectory():
    notebook_op = ExecuteFileOp(name="test",
                                pipeline_name="test-pipeline",
                                experiment_name="experiment-name",
                                notebook="foo/test.py",
                                cos_endpoint="http://testserver:32525",
                                cos_bucket="test_bucket",
                                cos_directory="test_directory",
                                cos_dependencies_archive="test_archive.tgz",
                                image="test/image:dev")
    assert "test.py" == notebook_op.notebook_name


def test_user_crio_volume_creation():
    notebook_op = ExecuteFileOp(name="test",
                                pipeline_name="test-pipeline",
                                experiment_name="experiment-name",
                                notebook="test_notebook.ipynb",
                                cos_endpoint="http://testserver:32525",
                                cos_bucket="test_bucket",
                                cos_directory="test_directory",
                                cos_dependencies_archive="test_archive.tgz",
                                image="test/image:dev",
                                emptydir_volume_size='20Gi')
    assert notebook_op.emptydir_volume_size == '20Gi'
    assert notebook_op.container_work_dir_root_path == '/opt/app-root/src/'
    assert notebook_op.container.volume_mounts.__len__() == 1
    # Environment variables: PYTHONPATH, ELYRA_RUN_NAME
    assert notebook_op.container.env.__len__() == 2, notebook_op.container.env


def test_override_bootstrap_url():
    notebook_op = ExecuteFileOp(name="test",
                                pipeline_name="test-pipeline",
                                experiment_name="experiment-name",
                                bootstrap_script_url="https://test.server.com/bootscript.py",
                                notebook="test_notebook.ipynb",
                                cos_endpoint="http://testserver:32525",
                                cos_bucket="test_bucket",
                                cos_directory="test_directory",
                                cos_dependencies_archive="test_archive.tgz",
                                image="test/image:dev")
    assert notebook_op.bootstrap_script_url == "https://test.server.com/bootscript.py"


def test_override_requirements_url():
    notebook_op = ExecuteFileOp(name="test",
                                pipeline_name="test-pipeline",
                                experiment_name="experiment-name",
                                requirements_url="https://test.server.com/requirements.py",
                                notebook="test_notebook.ipynb",
                                cos_endpoint="http://testserver:32525",
                                cos_bucket="test_bucket",
                                cos_directory="test_directory",
                                cos_dependencies_archive="test_archive.tgz",
                                image="test/image:dev")
    assert notebook_op.requirements_url == "https://test.server.com/requirements.py"


def test_construct_with_both_pipeline_inputs_and_outputs():
    notebook_op = ExecuteFileOp(name="test",
                                pipeline_name="test-pipeline",
                                experiment_name="experiment-name",
                                notebook="test_notebook.ipynb",
                                cos_endpoint="http://testserver:32525",
                                cos_bucket="test_bucket",
                                cos_directory="test_directory",
                                cos_dependencies_archive="test_archive.tgz",
                                pipeline_inputs=['test_input1.txt', 'test_input2.txt'],
                                pipeline_outputs=['test_output1.txt', 'test_output2.txt'],
                                image="test/image:dev")
    assert notebook_op.pipeline_inputs == ['test_input1.txt', 'test_input2.txt']
    assert notebook_op.pipeline_outputs == ['test_output1.txt', 'test_output2.txt']

    assert '--inputs "test_input1.txt;test_input2.txt"' in notebook_op.container.args[0]
    assert '--outputs "test_output1.txt;test_output2.txt"' in notebook_op.container.args[0]


def test_construct_wildcard_outputs():
    notebook_op = ExecuteFileOp(name="test",
                                pipeline_name="test-pipeline",
                                experiment_name="experiment-name",
                                notebook="test_notebook.ipynb",
                                cos_endpoint="http://testserver:32525",
                                cos_bucket="test_bucket",
                                cos_directory="test_directory",
                                cos_dependencies_archive="test_archive.tgz",
                                pipeline_inputs=['test_input1.txt', 'test_input2.txt'],
                                pipeline_outputs=['test_out*', 'foo.tar'],
                                image="test/image:dev")
    assert notebook_op.pipeline_inputs == ['test_input1.txt', 'test_input2.txt']
    assert notebook_op.pipeline_outputs == ['test_out*', 'foo.tar']

    assert '--inputs "test_input1.txt;test_input2.txt"' in notebook_op.container.args[0]
    assert '--outputs "test_out*;foo.tar"' in notebook_op.container.args[0]


def test_construct_with_only_pipeline_inputs():
    notebook_op = ExecuteFileOp(name="test",
                                pipeline_name="test-pipeline",
                                experiment_name="experiment-name",
                                notebook="test_notebook.ipynb",
                                cos_endpoint="http://testserver:32525",
                                cos_bucket="test_bucket",
                                cos_directory="test_directory",
                                cos_dependencies_archive="test_archive.tgz",
                                pipeline_inputs=['test_input1.txt', 'test,input2.txt'],
                                pipeline_outputs=[],
                                image="test/image:dev")
    assert notebook_op.pipeline_inputs == ['test_input1.txt', 'test,input2.txt']
    assert '--inputs "test_input1.txt;test,input2.txt"' in notebook_op.container.args[0]


def test_construct_with_bad_pipeline_inputs():
    with pytest.raises(ValueError) as error_info:
        ExecuteFileOp(name="test",
                      pipeline_name="test-pipeline",
                      experiment_name="experiment-name",
                      notebook="test_notebook.ipynb",
                      cos_endpoint="http://testserver:32525",
                      cos_bucket="test_bucket",
                      cos_directory="test_directory",
                      cos_dependencies_archive="test_archive.tgz",
                      pipeline_inputs=['test_input1.txt', 'test;input2.txt'],
                      pipeline_outputs=[],
                      image="test/image:dev")
    assert "Illegal character (;) found in filename 'test;input2.txt'." == str(error_info.value)


def test_construct_with_only_pipeline_outputs():
    notebook_op = ExecuteFileOp(name="test",
                                pipeline_name="test-pipeline",
                                experiment_name="experiment-name",
                                notebook="test_notebook.ipynb",
                                cos_endpoint="http://testserver:32525",
                                cos_bucket="test_bucket",
                                cos_directory="test_directory",
                                cos_dependencies_archive="test_archive.tgz",
                                pipeline_outputs=['test_output1.txt', 'test,output2.txt'],
                                pipeline_envs={},
                                image="test/image:dev")
    assert notebook_op.pipeline_outputs == ['test_output1.txt', 'test,output2.txt']
    assert '--outputs "test_output1.txt;test,output2.txt"' in notebook_op.container.args[0]


def test_construct_with_bad_pipeline_outputs():
    with pytest.raises(ValueError) as error_info:
        ExecuteFileOp(name="test",
                      pipeline_name="test-pipeline",
                      experiment_name="experiment-name",
                      notebook="test_notebook.ipynb",
                      cos_endpoint="http://testserver:32525",
                      cos_bucket="test_bucket",
                      cos_directory="test_directory",
                      cos_dependencies_archive="test_archive.tgz",
                      pipeline_outputs=['test_output1.txt', 'test;output2.txt'],
                      image="test/image:dev")
    assert "Illegal character (;) found in filename 'test;output2.txt'." == str(error_info.value)


def test_construct_with_env_variables_argo():
    notebook_op = ExecuteFileOp(name="test",
                                pipeline_name="test-pipeline",
                                experiment_name="experiment-name",
                                notebook="test_notebook.ipynb",
                                cos_endpoint="http://testserver:32525",
                                cos_bucket="test_bucket",
                                cos_directory="test_directory",
                                cos_dependencies_archive="test_archive.tgz",
                                pipeline_envs={"ENV_VAR_ONE": "1", "ENV_VAR_TWO": "2", "ENV_VAR_THREE": "3"},
                                image="test/image:dev")

    confirmation_names = ["ENV_VAR_ONE", "ENV_VAR_TWO", "ENV_VAR_THREE",
                          "ELYRA_RUN_NAME"]
    confirmation_values = ["1", "2", "3",
                           "{{workflow.annotations.pipelines.kubeflow.org/run_name}}"]
    for env_val in notebook_op.container.env:
        assert env_val.name in confirmation_names
        assert env_val.value in confirmation_values
        confirmation_names.remove(env_val.name)
        confirmation_values.remove(env_val.value)

    # Verify confirmation values have been drained.
    assert len(confirmation_names) == 0
    assert len(confirmation_values) == 0

    # same as before but explicitly specify the workflow engine type
    # as Argo
    notebook_op = ExecuteFileOp(name="test",
                                pipeline_name="test-pipeline",
                                experiment_name="experiment-name",
                                notebook="test_notebook.ipynb",
                                cos_endpoint="http://testserver:32525",
                                cos_bucket="test_bucket",
                                cos_directory="test_directory",
                                cos_dependencies_archive="test_archive.tgz",
                                pipeline_envs={"ENV_VAR_ONE": "1", "ENV_VAR_TWO": "2", "ENV_VAR_THREE": "3"},
                                image="test/image:dev",
                                workflow_engine="Argo")

    confirmation_names = ["ENV_VAR_ONE", "ENV_VAR_TWO", "ENV_VAR_THREE",
                          "ELYRA_RUN_NAME"]
    confirmation_values = ["1", "2", "3",
                           "{{workflow.annotations.pipelines.kubeflow.org/run_name}}"]
    for env_val in notebook_op.container.env:
        assert env_val.name in confirmation_names
        assert env_val.value in confirmation_values
        confirmation_names.remove(env_val.name)
        confirmation_values.remove(env_val.value)

    # Verify confirmation values have been drained.
    assert len(confirmation_names) == 0
    assert len(confirmation_values) == 0


def test_construct_with_env_variables_tekton():
    notebook_op = ExecuteFileOp(name="test",
                                pipeline_name="test-pipeline",
                                experiment_name="experiment-name",
                                notebook="test_notebook.ipynb",
                                cos_endpoint="http://testserver:32525",
                                cos_bucket="test_bucket",
                                cos_directory="test_directory",
                                cos_dependencies_archive="test_archive.tgz",
                                pipeline_envs={"ENV_VAR_ONE": "1", "ENV_VAR_TWO": "2", "ENV_VAR_THREE": "3"},
                                image="test/image:dev",
                                workflow_engine="Tekton")

    confirmation_names = ["ENV_VAR_ONE", "ENV_VAR_TWO", "ENV_VAR_THREE",
                          "ELYRA_RUN_NAME"]
    confirmation_values = ["1", "2", "3"]
    field_path = "metadata.annotations['pipelines.kubeflow.org/run_name']"
    for env_val in notebook_op.container.env:
        assert env_val.name in confirmation_names
        confirmation_names.remove(env_val.name)
        if env_val.name == 'ELYRA_RUN_NAME':
            assert env_val.value_from.field_ref.field_path == field_path, env_val.value_from.field_ref
        else:
            assert env_val.value in confirmation_values
            confirmation_values.remove(env_val.value)

    # Verify confirmation values have been drained.
    assert len(confirmation_names) == 0
    assert len(confirmation_values) == 0


def test_normalize_label_value():
    valid_middle_chars = '-_.'

    # test min length
    assert ExecuteFileOp._normalize_label_value(None) == ''
    assert ExecuteFileOp._normalize_label_value('') == ''
    # test max length (63)
    assert ExecuteFileOp._normalize_label_value('a' * 63) ==\
        'a' * 63
    assert ExecuteFileOp._normalize_label_value('a' * 64) ==\
        'a' * 63  # truncated
    # test first and last char
    assert ExecuteFileOp._normalize_label_value('1') == '1'
    assert ExecuteFileOp._normalize_label_value('22') == '22'
    assert ExecuteFileOp._normalize_label_value('3_3') == '3_3'
    assert ExecuteFileOp._normalize_label_value('4u4') == '4u4'
    assert ExecuteFileOp._normalize_label_value('5$5') == '5_5'

    # test first char
    for c in string.printable:
        if c in string.ascii_letters + string.digits:
            # first char is valid
            # no length violation
            assert ExecuteFileOp._normalize_label_value(c) == c
            assert ExecuteFileOp._normalize_label_value(c + 'B') == c + 'B'
            # max length
            assert ExecuteFileOp._normalize_label_value(c + 'B' * 62) == \
                   (c + 'B' * 62)
            # max length exceeded
            assert ExecuteFileOp._normalize_label_value(c + 'B' * 63) == \
                   (c + 'B' * 62)  # truncated
        else:
            # first char is invalid, e.g. '#a', and becomes the
            # second char, which might require replacement
            rv = c
            if c not in valid_middle_chars:
                rv = '_'
            # no length violation
            assert ExecuteFileOp._normalize_label_value(c) == 'a' + rv + 'a'
            assert ExecuteFileOp._normalize_label_value(c + 'B') == 'a' + rv + 'B'
            # max length
            assert ExecuteFileOp._normalize_label_value(c + 'B' * 62) == \
                   ('a' + rv + 'B' * 61)  # truncated
            # max length exceeded
            assert ExecuteFileOp._normalize_label_value(c + 'B' * 63) == \
                   ('a' + rv + 'B' * 61)  # truncated

    # test last char
    for c in string.printable:
        if c in string.ascii_letters + string.digits:
            # no length violation
            assert ExecuteFileOp._normalize_label_value('b' + c) == 'b' + c
            # max length
            assert ExecuteFileOp._normalize_label_value('b' * 62 + c) == \
                   ('b' * 62 + c)
            # max length exceeded
            assert ExecuteFileOp._normalize_label_value('b' * 63 + c) == \
                   ('b' * 63)
        else:
            # last char is invalid, e.g. 'a#', and requires
            # patching
            rv = c
            if c not in valid_middle_chars:
                rv = '_'
            # no length violation (char is appended)
            assert ExecuteFileOp._normalize_label_value('b' + c) == 'b' + rv + 'a'
            # max length (char is replaced)
            assert ExecuteFileOp._normalize_label_value('b' * 62 + c) == \
                   ('b' * 62 + 'a')
            # max length exceeded (no action required)
            assert ExecuteFileOp._normalize_label_value('b' * 63 + c) == \
                   ('b' * 63)

    # test first and last char
    for c in string.printable:
        if c in string.ascii_letters + string.digits:
            # no length violation
            assert ExecuteFileOp._normalize_label_value(c + 'b' + c) == \
                   c + 'b' + c  # nothing is modified
            # max length
            assert ExecuteFileOp._normalize_label_value(c + 'b' * 61 + c) == \
                   (c + 'b' * 61 + c)  # nothing is modified
            # max length exceeded
            assert ExecuteFileOp._normalize_label_value(c + 'b' * 62 + c) == \
                   c + 'b' * 62  # truncate only
        else:
            # first and last characters are invalid, e.g. '#a#'
            rv = c
            if c not in valid_middle_chars:
                rv = '_'
            # no length violation
            assert ExecuteFileOp._normalize_label_value(c + 'b' + c) ==\
                'a' + rv + 'b' + rv + 'a'
            # max length
            assert ExecuteFileOp._normalize_label_value(c + 'b' * 59 + c) == \
                   ('a' + rv + 'b' * 59 + rv + 'a')
            # max length exceeded after processing, scenario 1
            # resolved by adding char before first, replace last
            assert ExecuteFileOp._normalize_label_value(c + 'b' * 60 + c) == \
                   ('a' + rv + 'b' * 60 + 'a')
            # max length exceeded after processing, scenario 2
            # resolved by adding char before first, appending after last
            assert ExecuteFileOp._normalize_label_value(c + 'b' * 59 + c) == \
                   ('a' + rv + 'b' * 59 + rv + 'a')
            # max length exceeded before processing, scenario 1
            # resolved by adding char before first, truncating last
            assert ExecuteFileOp._normalize_label_value(c + 'b' * 62 + c) == \
                   ('a' + rv + 'b' * 61)
            # max length exceeded before processing, scenario 2
            # resolved by adding char before first, replacing last
            assert ExecuteFileOp._normalize_label_value(c + 'b' * 60 + c * 3) == \
                   ('a' + rv + 'b' * 60 + 'a')

    # test char in a position other than first and last
    # if invalid, the char is replaced with '_'
    for c in string.printable:
        if c in string.ascii_letters + string.digits + '-_.':
            assert ExecuteFileOp._normalize_label_value('A' + c + 'Z') ==\
                'A' + c + 'Z'
        else:
            assert ExecuteFileOp._normalize_label_value('A' + c + 'Z') == 'A_Z'

    # encore
    assert ExecuteFileOp._normalize_label_value(r'¯\_(ツ)_/¯') == 'a_________a'



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Any

from elyra.metadata.metadata import Metadata
from elyra.pipeline.processor import PipelineProcessorRegistry


class ComponentRegistryMetadata(Metadata):
    """
    This class contains methods to trigger cache updates on modification
    and deletion of component registry metadata instances.
    """

    def post_save(self, **kwargs: Any) -> None:
        processor_type = self.metadata.get('runtime')

        # Get processor instance and update its cache
        try:
            processor = PipelineProcessorRegistry.instance().get_processor(processor_type=processor_type)
            if processor.component_registry.caching_enabled:
                processor.component_registry.update_cache()
        except Exception:
            pass

    def post_delete(self, **kwargs: Any) -> None:
        processor_type = self.metadata.get('runtime')

        # Get processor instance and update its cache
        try:
            processor = PipelineProcessorRegistry.instance().get_processor(processor_type=processor_type)
            if processor.component_registry.caching_enabled:
                processor.component_registry.update_cache()
        except Exception:
            pass



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from datetime import datetime
from http.client import responses
import json
from typing import List
from typing import Optional

from jupyter_server.base.handlers import APIHandler
from jupyter_server.utils import url_path_join
from tornado import web

from elyra.pipeline.component import Component
from elyra.pipeline.component_registry import ComponentRegistry
from elyra.pipeline.parser import PipelineParser
from elyra.pipeline.processor import PipelineProcessorManager
from elyra.pipeline.validation import PipelineValidationManager
from elyra.util.http import HttpErrorMixin


class PipelineExportHandler(HttpErrorMixin, APIHandler):
    """Handler to expose REST API to export pipelines"""

    @web.authenticated
    async def get(self):
        msg_json = dict(title="Operation not supported.")
        self.set_header("Content-Type", 'application/json')
        self.finish(msg_json)

    @web.authenticated
    async def post(self, *args, **kwargs):
        self.log.debug("Pipeline Export handler now executing post request")

        parent = self.settings.get('elyra')
        payload = self.get_json_body()

        self.log.debug("JSON payload: %s", json.dumps(payload, indent=2, separators=(',', ': ')))

        pipeline_definition = payload['pipeline']
        pipeline_export_format = payload['export_format']
        pipeline_export_path = payload['export_path']
        pipeline_overwrite = payload['overwrite']

        response = await PipelineValidationManager.instance().validate(pipeline=pipeline_definition)
        self.log.debug(f"Validation checks completed. Results as follows: {response.to_json()}")

        if not response.has_fatal:
            pipeline = PipelineParser(root_dir=self.settings['server_root_dir'],
                                      parent=parent).parse(pipeline_definition)

            pipeline_exported_path = await PipelineProcessorManager.instance().export(
                pipeline,
                pipeline_export_format,
                pipeline_export_path,
                pipeline_overwrite
            )
            json_msg = json.dumps({"export_path": pipeline_export_path})
            self.set_status(201)
            self.set_header("Content-Type", 'application/json')
            location = url_path_join(
                self.base_url,
                'api',
                'contents',
                pipeline_exported_path
            )
            self.set_header('Location', location)
        else:
            json_msg = json.dumps({
                'reason': responses.get(400),
                'message': 'Errors found in pipeline',
                'timestamp': datetime.now().strftime('%Y-%m-%d %H:%M:%S'),
                'issues': response.to_json().get('issues')
            })
            self.set_status(400)

        self.set_header("Content-Type", 'application/json')
        self.finish(json_msg)


class PipelineSchedulerHandler(HttpErrorMixin, APIHandler):
    """Handler to expose method calls to execute pipelines as batch jobs"""

    @web.authenticated
    async def get(self):
        msg_json = dict(title="Operation not supported.")
        self.write(msg_json)
        self.flush()

    @web.authenticated
    async def post(self, *args, **kwargs):
        self.log.debug("Pipeline SchedulerHandler now executing post request")

        parent = self.settings.get('elyra')
        pipeline_definition = self.get_json_body()
        self.log.debug("JSON payload: %s", pipeline_definition)

        response = await PipelineValidationManager.instance().validate(pipeline=pipeline_definition)

        self.log.debug(f"Validation checks completed. Results as follows: {response.to_json()}")

        if not response.has_fatal:
            self.log.debug("Processing the pipeline submission and executing request")
            pipeline = PipelineParser(root_dir=self.settings['server_root_dir'],
                                      parent=parent).parse(pipeline_definition)
            response = await PipelineProcessorManager.instance().process(pipeline)
            json_msg = json.dumps(response.to_json())
            self.set_status(200)
        else:
            json_msg = json.dumps({
                'reason': responses.get(400),
                'message': 'Errors found in pipeline',
                'timestamp': datetime.now().strftime('%Y-%m-%d %H:%M:%S'),
                'issues': response.to_json().get('issues')
            })
            self.set_status(400)

        self.set_header("Content-Type", 'application/json')
        self.finish(json_msg)


class PipelineComponentHandler(HttpErrorMixin, APIHandler):
    """Handler to expose method calls to retrieve pipelines editor component configuration"""

    @web.authenticated
    async def get(self, processor):
        self.log.debug(f'Retrieving pipeline components for: {processor} runtime')

        if PipelineProcessorManager.instance().is_supported_runtime(processor) is False:
            raise web.HTTPError(400, f"Invalid processor name '{processor}'")

        components: List[Component] = await PipelineProcessorManager.instance().get_components(processor)
        palette_json = ComponentRegistry.to_canvas_palette(components=components)

        self.set_status(200)
        self.set_header("Content-Type", 'application/json')
        self.finish(palette_json)


class PipelineComponentPropertiesHandler(HttpErrorMixin, APIHandler):
    """Handler to expose method calls to retrieve pipeline component_id properties"""

    @web.authenticated
    async def get(self, processor, component_id):
        self.log.debug(f'Retrieving pipeline component properties for component: {component_id}')

        if PipelineProcessorManager.instance().is_supported_runtime(processor) is False:
            raise web.HTTPError(400, f"Invalid processor name '{processor}'")

        if not component_id:
            raise web.HTTPError(400, "Missing component ID")

        component: Optional[Component] = \
            await PipelineProcessorManager.instance().get_component(processor, component_id)
        properties_json = ComponentRegistry.to_canvas_properties(component)

        self.set_status(200)
        self.set_header("Content-Type", 'application/json')

        self.finish(properties_json)


class PipelineValidationHandler(HttpErrorMixin, APIHandler):
    """Handler to expose method calls to validate pipeline payloads for errors"""

    @web.authenticated
    async def get(self):
        msg_json = dict(title="GET requests are not supported.")
        self.write(msg_json)
        self.flush()

    @web.authenticated
    async def post(self):
        self.log.debug("Pipeline Validation Handler now executing post request")

        pipeline_definition = self.get_json_body()
        self.log.debug("Pipeline payload: %s", pipeline_definition)

        response = await PipelineValidationManager.instance().validate(pipeline=pipeline_definition)
        json_msg = response.to_json()

        self.set_status(200)
        self.set_header("Content-Type", 'application/json')
        self.finish(json_msg)



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Python Script with Environment Variables

This python script contains various environment variables to test the parser
functionality.
"""

import os

os.getenv("VAR1")
os.environ["VAR2"]
os.environ.get("VAR3")

print(os.environ['VAR4'])
print(os.getenv("VAR5", 'localhost'))

os.environ['VAR6'] = "value6"
print(os.environ.get('VAR7', 'value7'))
os.getenv('VAR8')

os.environ["VAR1"] = "newvalue"  # os.environ["VAR9"] = "value"

# os.getenv('VAR10')



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import ast
import logging
import sys
from typing import Any

"""Utility functions and classes used for metadata applications and classes."""

logging.basicConfig(level=logging.INFO, format='[%(levelname)1.1s %(asctime)s.%(msecs).03d] %(message)s')


class Option(object):
    """Represents the base option class.
    """
    cli_option = None
    name = None
    description = None
    default_value = None
    required = False
    value = None
    type = None  # Only used by SchemaProperty instances for now
    processed = False
    bad_value = None  # Contains error message string when bad value is encountered

    def __init__(self, cli_option, name=None, description=None, default_value=None, one_of=None,
                 required=False, type="string"):
        self.cli_option = cli_option
        self.name = name
        self.description = description
        self.default_value = default_value
        self.value = default_value
        self.one_of = one_of
        self.required = required
        self.type = type
        self.bad_value = None  # All options start as 'good'.  set_value() can set this to an error message

    def set_value(self, value):
        try:
            if self.type == 'array':
                self.value = Option.coerce_array_value(value)
            elif self.type == 'object':
                self.value = ast.literal_eval(value)
            elif self.type == 'integer':
                self.value = int(value)
            elif self.type == 'number':
                if "." in value:
                    self.value = float(value)
                else:
                    self.value = int(value)
            elif self.type == 'boolean':
                if isinstance(value, bool):
                    self.value = value
                elif str(value).lower() in ("true", "1"):
                    self.value = True
                elif str(value).lower() in ("false", "0"):
                    self.value = False
                else:
                    self.value = value  # let it take its course
            elif self.type == 'null':
                if str(value) in ("null", "None"):
                    self.value = None
                else:
                    self.value = value
            else:
                self.value = value
        except (ValueError, SyntaxError):
            self.handle_value_error(value)

    @staticmethod
    def coerce_array_value(value):
        new_value = value
        if value[0] != '[' and value[-1] != ']':  # attempt to coerce to list
            new_value = str(value.split(","))
        # The following assumes the array items should be strings and will break
        # non-quoted items like integers, numbers and booleans.  Its being left
        # here in case we want to support that scenario, which would likely mean
        # checking the entries to ensure they are expecting string values before
        # splitting.
        # elif value[0] == '[' and value[-1] == ']':
        #     # we have brackets.  If not internal quotes split within the brackets.
        #     # This handles the common (but invalid) "[item1,item2]" format.
        #     if value[1] not in ["'", '"'] and value[-2] not in ["'", '"']:
        #         new_value = str(value[1:-1].split(","))
        return ast.literal_eval(new_value)

    @staticmethod
    def get_article(type: str) -> str:
        vowels = ['a', 'e', 'i', 'o', 'u']   # we'll deal with 'y' as needed
        if type[0] in vowels:
            return "an"
        return "a"

    def get_format_hint(self) -> str:
        if self.one_of:
            msg = f"must be one of: {self.one_of}"
        elif self.type == 'array':
            msg = "\"['item1', 'item2']\" or \"item1,item2\""
        elif self.type == 'object':
            msg = "\"{'key1': 'value1', 'key2': 'value2'}\""
        elif self.type == 'integer':
            msg = "'n' where 'n' is an integer"
        elif self.type == 'number':
            msg = "'n.m' where 'n' and 'm' are integers"
        elif self.type == 'boolean':
            msg = "'true' or 'false'"
        elif self.type == 'null':
            msg = "'null' or 'None'"
        else:  # string
            msg = "sequence of characters"

        return msg

    def handle_value_error(self, value: Any) -> None:
        pre_amble = f"Parameter '{self.cli_option}' requires {Option.get_article(self.type)} {self.type} with format:"
        post_amble = f"and \"{value}\" was given.  Please try again with an appropriate value."
        self.bad_value = f"{pre_amble} {self.get_format_hint()} {post_amble}"

    def print_help(self):

        if isinstance(self, Flag):
            print(self.cli_option)
        else:
            option_entry = f"{self.cli_option}=<{self.type}>"
            required_entry = ""
            if self.required:
                required_entry = 'Required. '
            format_entry = f"Format: {self.get_format_hint()}"
            print(f"{option_entry} ({required_entry}{format_entry})")

        self.print_description()

    def print_description(self):
        print(f"\t{self.description}")


class CliOption(Option):
    """Represents a command-line option."""
    def __init__(self, cli_option, **kwargs):
        super().__init__(cli_option, **kwargs)


class Flag(Option):
    """Represents a command-line flag.  When present, the value used is `not default_value`."""
    def __init__(self, flag, **kwargs):
        super().__init__(flag, type="boolean", **kwargs)


class SchemaProperty(CliOption):
    """Represents the necessary information to handle a property from the schema.
       No validation is performed on corresponding instance values since the
       schema validation in the metadata service applies that.
       SchemaProperty instances are initialized from the corresponding property stanza
       from the schema
    """
    # Skip the following meta-properties when building the description.  We will already
    # have description and type and the others are difficult to display in a succinct manner.
    # Schema validation will still enforce these.
    skipped_meta_properties = ['description', 'type', 'items', 'additionalItems', 'properties'
                               'propertyNames', 'dependencies', 'examples', 'contains',
                               'additionalProperties', 'patternProperties']
    # Turn off the inclusion of meta-property information in the printed help messages  (Issue #837)
    print_meta_properties = False

    def __init__(self, name, schema_property):
        self.schema_property = schema_property
        cli_option = '--' + name
        type = schema_property.get('type')

        super().__init__(cli_option=cli_option,
                         name=name,
                         description=schema_property.get('description'),
                         default_value=schema_property.get('default'),
                         one_of=schema_property.get('enum'),
                         type=type)

    def print_description(self):

        additional_clause = ""
        if self.print_meta_properties:  # Only if enabled
            for meta_prop, value in self.schema_property.items():
                if meta_prop in self.skipped_meta_properties:
                    continue
                additional_clause = self._build_clause(additional_clause, meta_prop, value)

        print("\t{}{}".format(self.description, additional_clause))

    def _build_clause(self, additional_clause, meta_prop, value):
        if len(additional_clause) == 0:
            additional_clause = additional_clause + "; "
        else:
            additional_clause = additional_clause + ", "
        additional_clause = additional_clause + meta_prop + ": " + str(value)
        return additional_clause


class MetadataSchemaProperty(SchemaProperty):
    """Represents the property from the schema that resides in the Metadata stanza.
    """
    def __init__(self, name, schema_property):
        super().__init__(name, schema_property)


class AppBase(object):
    """Base class for application-level classes.  Provides logging, arguments handling,
       help methods, and anything common to its derived classes.
    """
    subcommands = {}
    description = None
    argv = []
    argv_mappings = {}  # Contains separation of argument name to value

    def __init__(self, **kwargs):
        self.argv = kwargs['argv']
        self._get_argv_mappings()
        self.log = logging.getLogger()  # setup logger so that metadata service logging is displayed

    def _get_argv_mappings(self):
        """Walk argv and build mapping from argument to value for later processing. """
        log_option = None
        for arg in self.argv:
            if '=' in arg:
                option, value = arg.split('=', 1)
            else:
                option, value = arg, None
            # Check for --debug or --log-level option.  if cound set, appropriate
            # log-level and skip.  Note this so we can alter self.argv after processing.
            if option == '--debug':
                log_option = arg
                logging.getLogger().setLevel(logging.DEBUG)
                continue
            elif option == '--log-level':
                log_option = arg
                logging.getLogger().setLevel(value)
                continue
            self.argv_mappings[option] = value
        if log_option:
            self.argv.remove(log_option)

    def log_and_exit(self, msg=None, exit_status=1, display_help=False):
        if msg:
            print(msg)
        if display_help:
            print()
            self.print_help()
        self.exit(exit_status)

    def get_subcommand(self):
        """Checks argv[0] to see if it matches one of the expected subcommands. If so,
           that item is removed from argv and that subcommand tuple (class, description)
           is returned.  If no an expected subcommand is not found (None, None) is returned.
        """
        if len(self.argv) > 0:
            arg = self.argv[0]
            if arg in self.subcommands.keys():
                subcommand = self.subcommands.get(arg)
                self._remove_argv_entry(arg)
                return subcommand

            if arg in ['--help', '-h']:
                self.log_and_exit(display_help=True)
            else:
                print("Subcommand '{}' is invalid.".format(self.argv[0]))
        return None

    def exit_no_subcommand(self):
        print("No subcommand specified. Must specify one of: %s" % list(self.subcommands))
        print()
        self.print_description()
        self.print_subcommands()
        self.exit(1)

    @staticmethod
    def schema_to_options(schema):
        """Takes a JSON schema and builds a list of SchemaProperty instances corresponding to each
           property in the schema.  There are two sections of properties, one that includes
           schema_name and display_name and another within the metadata container - which
           will be separated by class type - SchemaProperty vs. MetadataSchemaProperty.
        """
        options = {}
        properties = schema['properties']
        for name, value in properties.items():
            if name == 'schema_name':  # already have this option, skip
                continue
            if name != 'metadata':
                options[name] = SchemaProperty(name, value)
            else:  # process metadata properties...
                metadata_properties = properties['metadata']['properties']
                for md_name, md_value in metadata_properties.items():
                    options[md_name] = MetadataSchemaProperty(md_name, md_value)

        # Now set required-ness on MetadataProperties and top-level Properties
        required_props = properties['metadata'].get('required')
        for required in required_props:
            options.get(required).required = True

        required_props = schema.get('required')
        for required in required_props:
            # skip schema_name & metadata, already required, and metadata is not an option to be presented
            if required not in ['schema_name', 'metadata']:
                options.get(required).required = True
        return list(options.values())

    def process_cli_option(self, cli_option, check_help=False):
        """Check if the given option exists in the current arguments.  If found set its
           the Option instance's value to that of the argv.  Once processed, update the
           argv lists by removing the option.  If the option is a required property and
           is not in the argv lists or does not have a value, exit.
        """
        # if check_help is enabled, check the arguments for help options and
        # exit if found. This is only necessary when processing invidual options.
        if check_help and self.has_help():
            self.log_and_exit(display_help=True)

        if cli_option.processed:
            return
        option = cli_option.cli_option
        if option in self.argv_mappings.keys():
            if isinstance(cli_option, Flag):  # flags set their value opposite their default
                cli_option.value = not cli_option.default_value
            else:  # this is a regular option, just set value
                cli_option.set_value(self.argv_mappings.get(option))
                if cli_option.bad_value:
                    self.log_and_exit(cli_option.bad_value, display_help=True)
                if cli_option.required:
                    if not cli_option.value:
                        self.log_and_exit("Parameter '{}' requires a value.".
                                          format(cli_option.cli_option), display_help=True)
                    elif cli_option.one_of:  # ensure value is in set
                        if cli_option.value not in cli_option.one_of:
                            self.log_and_exit("Parameter '{}' requires one of the following values: {}".
                                              format(cli_option.cli_option, cli_option.one_of), display_help=True)
            self._remove_argv_entry(option)
        elif cli_option.required and cli_option.value is None:
            if cli_option.one_of is None:
                self.log_and_exit("'{}' is a required parameter.".
                                  format(cli_option.cli_option), display_help=True)
            else:
                self.log_and_exit("'{}' is a required parameter and must be one of the following values: {}.".
                                  format(cli_option.cli_option, cli_option.one_of), display_help=True)

        cli_option.processed = True

    def process_cli_options(self, cli_options):
        """For each Option instance in the list, process it according to the argv lists.
           After traversal, if arguments still remain, log help and exit.
        """
        # Since we're down to processing options (no subcommands), scan the arguments
        # for help entries and, if found, exit with the help message.
        if self.has_help():
            self.log_and_exit(display_help=True)

        for option in cli_options:
            self.process_cli_option(option)

        # Check if there are still unprocessed arguments.  If so, and fail_unexpected is true,
        # log and exit, else issue warning and continue.
        if len(self.argv) > 0:
            msg = "The following arguments were unexpected: {}".format(self.argv)
            self.log_and_exit(msg, display_help=True)

    def has_help(self):
        """Checks the arguments to see if any match the help options.
           We do this by converting two lists to sets and checking if
           there's an intersection.
        """
        helps = {'--help', '-h'}
        args = set(self.argv_mappings.keys())
        help_list = list(helps & args)
        return len(help_list) > 0

    def _remove_argv_entry(self, cli_option):
        """Removes the argument entry corresponding to cli_option in both
           self.argv and self.argv_mappings
        """
        # build the argv entry from the mappings since it must be located with name=value
        if cli_option not in self.argv_mappings.keys():
            self.log_and_exit("Can't find option '{}' in argv!".format(cli_option))

        entry = cli_option
        value = self.argv_mappings.get(cli_option)
        if value:
            entry = entry + '=' + value
        self.argv.remove(entry)
        self.argv_mappings.pop(cli_option)

    def print_help(self):
        self.print_description()

    def print_description(self):
        print(self.description)

    def print_subcommands(self):
        print()
        print("Subcommands")
        print("-----------")
        print("Subcommands are launched as `elyra-metadata cmd [args]`. For information on")
        print("using subcommand 'cmd', run: `elyra-metadata cmd -h`.")
        print()
        for subcommand, desc in self.subcommands.items():
            print(subcommand)
            print("    {}".format(desc[1]))

    def exit(self, status):
        sys.exit(status)



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
from typing import Any
from typing import Type
from typing import TypeVar

from ipython_genutils.importstring import import_item

from elyra.metadata.schema import SchemaManager

# Setup forward reference for type hint on return from class factory method.  See
# https://stackoverflow.com/questions/39205527/can-you-annotate-return-type-when-value-is-instance-of-cls/39205612#39205612
M = TypeVar('M', bound='Metadata')


class Metadata(object):
    name = None
    resource = None
    display_name = None
    schema_name = None
    metadata = {}
    reason = None

    def __init__(self, **kwargs: Any) -> None:
        self.name = kwargs.get('name')
        self.display_name = kwargs.get('display_name')
        self.schema_name = kwargs.get('schema_name')
        self.metadata = kwargs.get('metadata', {})
        self.resource = kwargs.get('resource')
        self.reason = kwargs.get('reason')

    def post_load(self, **kwargs: Any) -> None:
        """Called by MetadataManager after fetching the instance.

        :param kwargs: additional arguments
        """
        pass

    def pre_save(self, **kwargs: Any) -> None:
        """Called by MetadataManager prior to saving the instance.

        :param kwargs: additional arguments
        Keyword Args:
            for_update (bool): indicates if this save operation if for update (True) or create (False)
        """
        pass

    def post_save(self, **kwargs: Any) -> None:
        """Called by MetadataManager following the save of the instance.

        :param kwargs: additional arguments
        Keyword Args:
            for_update (bool): indicates if this save operation if for update (True) or create (False)

        Note: Since exceptions thrown from this method can adversely affect the operation,
        implementations are advised to trap all exceptions unless the operation's
        rollback is warranted.
        """
        pass

    def pre_delete(self, **kwargs: Any) -> None:
        """Called by MetadataManager prior to deleting the instance."""
        pass

    def post_delete(self, **kwargs: Any) -> None:
        """Called by MetadataManager following the deletion of the instance.

        Note: Since exceptions thrown from this method can adversely affect the operation,
        implementations are advised to trap all exceptions unless the operation's
        rollback is warranted.
        """
        pass

    @classmethod
    def from_dict(cls: Type[M], schemaspace: str, metadata_dict: dict) -> M:
        """Creates an appropriate instance of Metadata from a dictionary instance """

        # Get the schema and look for metadata_class entry and use that, else Metadata.
        metadata_class_name = 'elyra.metadata.metadata.Metadata'
        schema_name = metadata_dict.get('schema_name')
        if schema_name:
            schema = SchemaManager.instance().get_schema(schemaspace, schema_name)
            metadata_class_name = schema.get('metadata_class_name', metadata_class_name)
        metadata_class = import_item(metadata_class_name)
        try:
            instance = metadata_class(**metadata_dict)
            if not isinstance(instance, Metadata):
                raise ValueError("The metadata_class_name ('{}') for schema '{}' must be a subclass of '{}'!".
                                 format(metadata_class_name, schema_name, cls.__name__))
        except TypeError as te:
            raise ValueError("The metadata_class_name ('{}') for schema '{}' must be a subclass of '{}'!".
                             format(metadata_class_name, schema_name, cls.__name__)) from te
        return instance

    def to_dict(self, trim: bool = False) -> dict:
        # Exclude resource, and reason only if trim is True since we don't want to persist that information.
        #  Method prepare_write will be used to remove name prior to writes.
        d = dict(name=self.name, display_name=self.display_name, metadata=self.metadata, schema_name=self.schema_name)
        if not trim:
            if self.resource:
                d['resource'] = self.resource
            if self.reason:
                d['reason'] = self.reason

        return d

    def to_json(self, trim: bool = False) -> str:
        return json.dumps(self.to_dict(trim=trim), indent=2)

    def prepare_write(self) -> dict:
        """Prepares this instance for storage, stripping name, reason, and resource and converting to a dict"""
        prepared = self.to_dict(trim=True)  # we should also trim 'name' when storing
        prepared.pop('name', None)
        return prepared

    def __repr__(self):
        return self.to_json()



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import sys

from glob import glob
from setuptools import setup, find_packages

long_desc = """
            Elyra is a set of AI centric extensions to JupyterLab. It aims to help data scientists,
            machine learning engineers and AI developer’s through the model development life cycle complexities.
            """

here = os.path.abspath(os.path.dirname(__file__))

version_ns = {}
with open(os.path.join(here, 'elyra', '_version.py')) as f:
    exec(f.read(), {}, version_ns)

npm_packages_path = "./dist/*.tgz"
auto_jupyter_notebook_extension_path = "./etc/config/jupyter_notebook_config.d/*.json"
auto_jupyter_server_extension_path = "./etc/config/jupyter_server_config.d/*.json"
component_registry_path = './etc/config/components/*.json'
components_kfp_path = './etc/config/components/kfp/*.yaml'
components_airflow_path = './etc/config/components/airflow/*.py'
metadata_path_runtime_image = './etc/config/metadata/runtime-images/*.json'
metadata_path_registries = './etc/config/metadata/component-registries/*.json'
settings_path = './etc/config/settings/*.json'

runtime_extras = {
    'kfp-tekton': ['kfp-tekton~=0.8.1',]
}
runtime_extras['all'] = list(set(sum(runtime_extras.values(), [])))

setup_args = dict(
    name="elyra-server",
    version=version_ns['__version__'],
    url="https://github.com/elyra-ai/elyra",
    description="Elyra provides AI Centric extensions to JupyterLab",
    long_description=long_desc,
    author="Elyra Maintainers",
    license="Apache License Version 2.0",
    data_files=[('etc/jupyter/jupyter_notebook_config.d', glob(auto_jupyter_notebook_extension_path)),
                ('etc/jupyter/jupyter_server_config.d', glob(auto_jupyter_server_extension_path)),
                ('share/jupyter/metadata/runtime-images', glob(metadata_path_runtime_image)),
                ('share/jupyter/metadata/component-registries', glob(metadata_path_registries)),
                ('share/jupyter/components', glob(component_registry_path)),
                ('share/jupyter/components/kfp/', glob(components_kfp_path)),
                ('share/jupyter/components/airflow/', glob(components_airflow_path)),
                ('share/jupyter/lab/settings', glob(settings_path))],
    packages=find_packages(),
    install_requires=[
        'autopep8>=1.5.0,<1.5.6',
        'click>=7.1.1,<8', #Required bykfp 1.6.3
        'colorama',
        'entrypoints>=0.3',
        'jinja2>=2.11',
        'jsonschema>=3.2.0',
        'jupyter_core>=4.0,<5.0',
        'jupyter_client>=6.1.7',
        'jupyter_server>=1.7.0',
        'jupyterlab>=3.0.17',
        'jupyterlab-git~=0.32',
        'jupyterlab-lsp>=3.8.0',
        'jupyter-resource-usage>=0.5.1',
        'minio>=5.0.7,<7.0.0',
        'nbclient>=0.5.1',
        'nbconvert>=5.6.1',
        'nbdime~=3.1',
        'nbformat>=5.1.2',
        'networkx>=2.5.1',
        'papermill>=2.1.3',
        'python-lsp-server[all]>=1.1.0',
        'pyyaml>=5.3.1,<6.0',
        'requests>=2.25.1,<3.0',
        'rfc3986-validator>=0.1.1',
        'tornado>=6.1.0',
        'traitlets>=4.3.2',
        'urllib3>=1.26.5',
        'watchdog>=2.1.3',
        'websocket-client',
        'yaspin',
        # KFP runtime dependencies
        'kfp>=1.6.3<2.0,!=1.7.2',
        # Airflow runtime dependencies
        'pygithub',
        'black',
    ],
    extras_require={
        'test': ['pytest', 'pytest-tornasync'],
        **runtime_extras
    },
    include_package_data=True,
    classifiers=(
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ),
    entry_points={
        'console_scripts': [
            'elyra-metadata = elyra.metadata.metadata_app:MetadataApp.main',
            'elyra-pipeline = elyra.cli.pipeline_app:pipeline',
            'jupyter-elyra = elyra.elyra_app:launch_instance'
        ],
        'metadata.schemaspaces': [
            'runtimes = elyra.metadata.schemaspaces:Runtimes',
            'runtimes-images = elyra.metadata.schemaspaces:RuntimeImages',
            'code-snippets = elyra.metadata.schemaspaces:CodeSnippets',
            'component-registries = elyra.metadata.schemaspaces:ComponentRegistries',
            'metadata-tests = elyra.tests.metadata.test_utils:MetadataTestSchemaspace'
        ],
        'metadata.schemas_providers': [
            'runtimes = elyra.metadata.schemasproviders:RuntimesSchemas',
            'runtimes-images = elyra.metadata.schemasproviders:RuntimeImagesSchemas',
            'code-snippets = elyra.metadata.schemasproviders:CodeSnippetsSchemas',
            'component-registries = elyra.metadata.schemasproviders:ComponentRegistriesSchemas',
            'metadata-tests = elyra.tests.metadata.test_utils:MetadataTestSchemasProvider'
        ],
        'elyra.pipeline.processors': [
            'local = elyra.pipeline.local.processor_local:LocalPipelineProcessor',
            'airflow = elyra.pipeline.airflow.processor_airflow:AirflowPipelineProcessor',
            'kfp = elyra.pipeline.kfp.processor_kfp:KfpPipelineProcessor'
        ],
        'papermill.engine': [
            'ElyraEngine = elyra.pipeline.elyra_engine:ElyraEngine',
        ]
    },
)


if "--dev" not in sys.argv:
    setup_args["data_files"].append(('share/jupyter/lab/extensions', glob(npm_packages_path)))
else:
    sys.argv.remove("--dev")

if __name__ == '__main__':
    setup(**setup_args)
